import os
import threading

from platypush.context import get_bus
from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.media import (
    MediaPlayEvent,
    MediaPlayRequestEvent,
    MediaPauseEvent,
    MediaStopEvent,
    NewPlayingMediaEvent,
    MediaSeekEvent,
    MediaResumeEvent,
)

from platypush.plugins import action


class MediaMpvPlugin(MediaPlugin):
    """
    Plugin to control MPV instances

    Requires:

        * **python-mpv** (``pip install python-mpv``)
        * **mpv** executable on your system
    """

    _default_mpv_args = {
        'ytdl': True,
        'start_event_thread': True,
    }

    def __init__(self, args=None, *argv, **kwargs):
        """
        Create the MPV wrapper.

        :param args: Default arguments that will be passed to the mpv executable
            as a key-value dict (names without the `--` prefix). See `man mpv`
            for available options.
        :type args: dict[str, str]
        """

        super().__init__(*argv, **kwargs)

        self.args = self._default_mpv_args
        if args:
            # noinspection PyTypeChecker
            self.args.update(args)

        self._player = None
        self._playback_rebounce_event = threading.Event()
        self._on_stop_callbacks = []

    def _init_mpv(self, args=None):
        import mpv

        mpv_args = self.args.copy()
        if args:
            mpv_args.update(args)

        for k, v in self._env.items():
            os.environ[k] = v

        self._player = mpv.MPV(**mpv_args)
        # noinspection PyProtectedMember
        self._player._event_callbacks += [self._event_callback()]

    @staticmethod
    def _post_event(evt_type, **evt):
        bus = get_bus()
        bus.post(evt_type(player='local', plugin='media.mpv', **evt))

    def _event_callback(self):
        def callback(event):
            from mpv import (
                MpvEvent,
                MpvEventID as Event,
                MpvEventEndFile as EndFile,
            )

            self.logger.info('Received mpv event: {}'.format(event))

            if isinstance(event, MpvEvent):
                event = event.as_dict()

            evt = event.get('event_id')
            if not evt:
                return

            if (
                evt == Event.FILE_LOADED or evt == Event.START_FILE
            ) and self._get_current_resource():
                self._playback_rebounce_event.set()
                self._post_event(
                    NewPlayingMediaEvent,
                    resource=self._get_current_resource(),
                    title=self._player.filename,
                )
            elif evt == Event.PLAYBACK_RESTART:
                self._playback_rebounce_event.set()
                self._post_event(
                    MediaPlayEvent,
                    resource=self._get_current_resource(),
                    title=self._player.filename,
                )
            elif evt == Event.PAUSE:
                self._post_event(
                    MediaPauseEvent,
                    resource=self._get_current_resource(),
                    title=self._player.filename,
                )
            elif evt == Event.UNPAUSE:
                self._post_event(
                    MediaResumeEvent,
                    resource=self._get_current_resource(),
                    title=self._player.filename,
                )
            elif (
                evt == Event.SHUTDOWN
                or evt == Event.IDLE
                or (
                    evt == Event.END_FILE
                    and event.get('event', {}).get('reason')
                    in [EndFile.EOF, EndFile.ABORTED, EndFile.QUIT]
                )
            ):
                playback_rebounced = self._playback_rebounce_event.wait(timeout=0.5)
                if playback_rebounced:
                    self._playback_rebounce_event.clear()
                    return

                self._player = None
                self._post_event(MediaStopEvent)

                for cbk in self._on_stop_callbacks:
                    cbk()
            elif evt == Event.SEEK:
                self._post_event(MediaSeekEvent, position=self._player.playback_time)

        return callback

    @action
    def execute(self, cmd, **args):
        """
        Execute a raw mpv command.
        """
        if not self._player:
            return None, 'No mpv instance is running'
        return self._player.command(cmd, *args)

    @action
    def play(self, resource, subtitles=None, **args):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
        :type resource: str

        :param subtitles: Path to optional subtitle file
        :type subtitles: str

        :param args: Extra runtime arguments that will be passed to the
            mpv executable as a key-value dict (keys without `--` prefix)
        :type args: dict[str,str]
        """

        self._post_event(MediaPlayRequestEvent, resource=resource)
        self._init_mpv(args)

        resource = self._get_resource(resource)
        if resource.startswith('file://'):
            resource = resource[7:]

        assert self._player, 'The player is not ready'
        self._player.play(resource)
        if self.volume:
            self.set_volume(volume=self.volume)
        if subtitles:
            self.add_subtitles(subtitles)

        return self.status()

    @action
    def pause(self):
        """Toggle the paused state"""
        if not self._player:
            return None, 'No mpv instance is running'

        self._player.pause = not self._player.pause
        return self.status()

    @action
    def quit(self):
        """Stop and quit the player"""
        if not self._player:
            return None, 'No mpv instance is running'

        self._player.quit()
        self._player.terminate()
        self._player = None
        return {'state': PlayerState.STOP.value}

    @action
    def stop(self):
        """Stop and quit the player"""
        return self.quit()

    @action
    def voldown(self, step=10.0):
        """Volume down by (default: 10)%"""
        if not self._player:
            return None, 'No mpv instance is running'
        return self.set_volume(self._player.volume - step)

    @action
    def volup(self, step=10.0):
        """Volume up by (default: 10)%"""
        if not self._player:
            return None, 'No mpv instance is running'
        return self.set_volume(self._player.volume + step)

    @action
    def set_volume(self, volume):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        :type volume: float
        """
        if not self._player:
            return None, 'No mpv instance is running'

        volume = max(0, min([self._player.volume_max, volume]))
        self._player.volume = volume
        return self.status()

    @action
    def seek(self, position):
        """
        Seek backward/forward by the specified number of seconds

        :param position: Number of seconds relative to the current cursor
        :type position: int
        """
        if not self._player:
            return None, 'No mpv instance is running'
        if not self._player.seekable:
            return None, 'The resource is not seekable'
        pos = min(self._player.time_pos + self._player.time_remaining, max(0, position))
        self._player.time_pos = pos
        return self.status()

    @action
    def back(self, offset=30.0):
        """Back by (default: 30) seconds"""
        if not self._player:
            return None, 'No mpv instance is running'
        if not self._player.seekable:
            return None, 'The resource is not seekable'
        pos = max(0, self._player.time_pos - offset)
        return self.seek(pos)

    @action
    def forward(self, offset=30.0):
        """Forward by (default: 30) seconds"""
        if not self._player:
            return None, 'No mpv instance is running'
        if not self._player.seekable:
            return None, 'The resource is not seekable'
        pos = min(
            self._player.time_pos + self._player.time_remaining,
            self._player.time_pos + offset,
        )
        return self.seek(pos)

    @action
    def next(self):
        """Play the next item in the queue"""
        if not self._player:
            return None, 'No mpv instance is running'
        self._player.playlist_next()

    @action
    def prev(self):
        """Play the previous item in the queue"""
        if not self._player:
            return None, 'No mpv instance is running'
        self._player.playlist_prev()

    @action
    def toggle_subtitles(self, visible=None):
        """Toggle the subtitles visibility"""
        return self.toggle_property('sub_visibility')

    @action
    def add_subtitles(self, filename):
        """Add a subtitles file"""
        return self._player.sub_add(filename)

    @action
    def toggle_fullscreen(self):
        """Toggle the fullscreen mode"""
        return self.toggle_property('fullscreen')

    # noinspection PyShadowingBuiltins
    @action
    def toggle_property(self, property):
        """
        Toggle or sets the value of an mpv property (e.g. fullscreen,
        sub_visibility etc.). See ``man mpv`` for a full list of properties

        :param property: Property to toggle
        """
        if not self._player:
            return None, 'No mpv instance is running'

        if not hasattr(self._player, property):
            self.logger.warning('No such mpv property: {}'.format(property))

        value = not getattr(self._player, property)
        setattr(self._player, property, value)
        return {property: value}

    # noinspection PyShadowingBuiltins
    @action
    def get_property(self, property):
        """
        Get a player property (e.g. pause, fullscreen etc.). See
        ``man mpv`` for a full list of the available properties
        """
        if not self._player:
            return None, 'No mpv instance is running'
        return getattr(self._player, property)

    @action
    def set_property(self, **props):
        """
        Set the value of an mpv property (e.g. fullscreen, sub_visibility
        etc.). See ``man mpv`` for a full list of properties

        :param props: Key-value args for the properties to set
        :type props: dict
        """
        if not self._player:
            return None, 'No mpv instance is running'

        for k, v in props.items():
            setattr(self._player, k, v)
        return props

    @action
    def set_subtitles(self, filename, *args, **kwargs):
        """Sets media subtitles from filename"""
        # noinspection PyTypeChecker
        return self.set_property(subfile=filename, sub_visibility=True)

    @action
    def remove_subtitles(self, sub_id=None):
        """Removes (hides) the subtitles"""
        if not self._player:
            return None, 'No mpv instance is running'
        if sub_id:
            return self._player.sub_remove(sub_id)
        self._player.sub_visibility = False

    @action
    def is_playing(self):
        """
        :returns: True if it's playing, False otherwise
        """
        if not self._player:
            return False
        return not self._player.pause

    @action
    def load(self, resource, **args):
        """
        Load/queue a resource/video to the player
        """
        if not self._player:
            return self.play(resource, **args)
        return self._player.loadfile(resource, mode='append-play')

    @action
    def mute(self):
        """Toggle mute state"""
        if not self._player:
            return None, 'No mpv instance is running'
        mute = not self._player.mute
        self._player.mute = mute
        return {'muted': mute}

    @action
    def set_position(self, position):
        """
        Seek backward/forward to the specified absolute position (same as ``seek``)
        """
        return self.seek(position)

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example::

            output = {
                "filename": "filename or stream URL",
                "state": "play"  # or "stop" or "pause"
            }
        """
        if not self._player or not hasattr(self._player, 'pause'):
            return {'state': PlayerState.STOP.value}

        return {
            'audio_channels': getattr(self._player, 'audio_channels', None),
            'audio_codec': getattr(self._player, 'audio_codec_name', None),
            'delay': getattr(self._player, 'delay', None),
            'duration': getattr(self._player, 'playback_time', 0)
            + getattr(self._player, 'playtime_remaining', 0)
            if getattr(self._player, 'playtime_remaining', None)
            else None,
            'filename': getattr(self._player, 'filename', None),
            'file_size': getattr(self._player, 'file_size', None),
            'fullscreen': getattr(self._player, 'fs', None),
            'mute': getattr(self._player, 'mute', None),
            'name': getattr(self._player, 'name', None),
            'pause': getattr(self._player, 'pause', None),
            'percent_pos': getattr(self._player, 'percent_pos', None),
            'position': getattr(self._player, 'playback_time', None),
            'seekable': getattr(self._player, 'seekable', None),
            'state': (
                PlayerState.PAUSE.value
                if self._player.pause
                else PlayerState.PLAY.value
            ),
            'title': getattr(self._player, 'media_title', None)
            or getattr(self._player, 'filename', None),
            'url': self._get_current_resource(),
            'video_codec': getattr(self._player, 'video_codec', None),
            'video_format': getattr(self._player, 'video_format', None),
            'volume': getattr(self._player, 'volume', None),
            'volume_max': getattr(self._player, 'volume_max', None),
            'width': getattr(self._player, 'width', None),
        }

    def on_stop(self, callback):
        self._on_stop_callbacks.append(callback)

    def _get_current_resource(self):
        if not self._player or not self._player.stream_path:
            return

        return (
            'file://' if os.path.isfile(self._player.stream_path) else ''
        ) + self._player.stream_path

    def _get_resource(self, resource):
        if self._is_youtube_resource(resource):
            return resource  # mpv can handle YouTube streaming natively

        return super()._get_resource(resource)


# vim:sw=4:ts=4:et:



from platypush.message.event import Event


class ClipboardEvent(Event):
    def __init__(self, text: str, *args, **kwargs):
        super().__init__(*args, text=text, **kwargs)


# vim:sw=4:ts=4:et:



import os
from typing import Optional, Union, Sequence

import requests
from requests.auth import HTTPBasicAuth

from platypush.plugins import action
from platypush.plugins.mail import MailOutPlugin


class MailgunPlugin(MailOutPlugin):
    """
    Mailgun integration.
    """
    def __init__(self, api_key: str, api_base_url: str = 'https://api.mailgun.net/v3', **kwargs):
        """
        :param api_key: Mailgun API secret key.
        :param api_base_url: Use ``https://api.eu.mailgun.net/v3`` if you are using an EU account.
        """
        super().__init__(**kwargs)
        self._api_key = api_key
        self._api_base_url = api_base_url

    def send_message(self, *_, **__):
        pass

    @action
    def send(
            self, domain: str, to: Union[str, Sequence[str]], from_: Optional[str] = None,
            cc: Optional[Union[str, Sequence[str]]] = None, bcc: Optional[Union[str, Sequence[str]]] = None,
            subject: str = '', body: str = '', body_type: str = 'plain', attachments: Optional[Sequence[str]] = None,
            **kwargs
     ):
        """
        Send an email through Mailgun.

        :param domain: From which registered domain the email(s) should be sent.
        :param to: Receiver(s), as comma-separated strings or list.
        :param from_: Sender email address (``from`` is also supported outside of Python contexts).
        :param cc: Carbon-copy addresses, as comma-separated strings or list
        :param bcc: Blind carbon-copy addresses, as comma-separated strings or list
        :param subject: Mail subject.
        :param body: Mail body.
        :param body_type: Mail body type - ``text`` or ``html``.
        :param attachments: List of attachment files to send.
        """
        from_ = from_ or kwargs.pop('from', None)
        rs = requests.post(
            f'{self._api_base_url}/{domain}/messages',
            auth=HTTPBasicAuth('api', self._api_key),
            data={
                'to': ', '.join([to] if isinstance(to, str) else to),
                'subject': subject,
                **{'html' if body_type == 'html' else 'text': body},
                **({'from': from_} if from_ else {}),
                **({'cc': ', '.join([cc] if isinstance(cc, str) else cc)} if cc else {}),
                **({'bcc': ', '.join([bcc] if isinstance(bcc, str) else bcc)} if bcc else {}),
            },
            files=[os.path.expanduser(attachment) for attachment in (attachments or [])]
        )

        rs.raise_for_status()


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.plugins import action
from platypush.plugins.http.request import HttpRequestPlugin
from platypush.plugins.weather import WeatherPlugin


class WeatherOpenweathermapPlugin(HttpRequestPlugin, WeatherPlugin):
    """
    OpenWeatherMap plugin. This is the advised plugin to use for weather forecasts since Darksky has officially
    shut down their API.

    You'll need an API token from `OpenWeatherMap <https://openweathermap.org/api>`_ in order to use this API.
    """
    base_url = 'https://api.openweathermap.org/data/2.5/weather'

    def __init__(self, token: str, location: Optional[str] = None, city_id: Optional[int] = None,
                 lat: Optional[float] = None, long: Optional[float] = None,
                 zip_code: Optional[str] = None, units: str = 'metric', **kwargs):
        """
        :param token: OpenWeatherMap API token.
        :param location: If set, then this location will be used by default for weather lookup. If multiple locations
            share the same name you can disambiguate by specifying the country code as well - e.g. ``London,GB``.
        :param city_id: If set, then this city ID will be used by default for weather lookup. The full list of city IDs
            is available `here <https://bulk.openweathermap.org/sample/>`_.
        :param lat: If lat/long are set, then the weather by default will be retrieved for the specified geo location.
        :param long: If lat/long are set, then the weather by default will be retrieved for the specified geo location.
        :param zip_code: If set, then this ZIP code (should be in the form ``zip,country_code``) will be used by default
            for weather lookup.
        :param units: Supported: ``metric`` (default), ``standard`` and ``imperial``.
        """
        super().__init__(method='get', output='json', **kwargs)
        self._token = token
        self._location_query = None
        self._location_query = self._get_location_query(location=location, city_id=city_id, lat=lat, long=long,
                                                        zip_code=zip_code)
        self.units = units

    def _get_location_query(self, location: Optional[str] = None, city_id: Optional[int] = None,
                            lat: Optional[float] = None, long: Optional[float] = None,
                            zip_code: Optional[str] = None) -> dict:
        if city_id:
            return {'id': city_id}
        if lat and long:
            return {'lat': lat, 'lon': long}
        if zip_code:
            return {'zip': zip_code}
        if location:
            return {'q': location}

        assert self._location_query, 'Specify either location, city_id or lat/long'
        return self._location_query

    @action
    def get(self, url, **kwargs):
        kwargs['params'] = {
            'appid': self._token,
            **kwargs.get('params', {}),
        }

        return super().get(url, **kwargs)

    @staticmethod
    def _convert_percentage(perc: Optional[float]) -> Optional[float]:
        if perc is None:
            return
        return perc / 100.

    @staticmethod
    def _m_to_km(m: Optional[float]) -> Optional[float]:
        if m is None:
            return
        return m / 1000.

    @staticmethod
    def _get_precip_type(response: dict) -> Optional[str]:
        if response.get('snow', {}).get('1h', 0) > 0:
            return 'snow'
        if response.get('rain', {}).get('1h', 0) > 0:
            return 'rain'
        return

    @classmethod
    def _convert_weather_response(cls, response: dict) -> dict:
        return {
            'time': response.get('dt'),
            'summary': response.get('weather', [{'main': 'Unknown'}])[0].get('main', 'Unknown'),
            'icon': response.get('weather', [{'icon': 'unknown'}])[0].get('icon', 'unknown'),
            'precipIntensity': response.get('rain', response.get('snow', {})).get('1h', 0),
            'precipType': cls._get_precip_type(response),
            'temperature': response.get('main', {}).get('temp'),
            'apparentTemperature': response.get('main', {}).get('feels_like'),
            'humidity': cls._convert_percentage(response.get('main', {}).get('humidity')),
            'pressure': response.get('main', {}).get('pressure'),
            'windSpeed': response.get('wind', {}).get('speed'),
            'windDirection': response.get('wind', {}).get('deg'),
            'windGust': response.get('wind', {}).get('gust'),
            'cloudCover': cls._convert_percentage(response.get('clouds', {}).get('all')),
            'visibility': cls._m_to_km(response.get('visibility')),
            'sunrise': response.get('sys', {}).get('sunrise'),
            'sunset': response.get('sys', {}).get('sunset'),
        }

    @action
    def get_current_weather(self, *, location: Optional[str] = None, city_id: Optional[int] = None,
                            lat: Optional[float] = None, long: Optional[float] = None, zip_code: Optional[str] = None,
                            units: Optional[str] = None, **kwargs) -> dict:
        """
        Returns the current weather.

        :param location: Override the ``location`` configuration value.
        :param city_id: Override the ``city_id`` configuration value.
        :param lat: Override the ``lat`` configuration value.
        :param long: Override the ``long`` configuration value.
        :param zip_code: Override the ``zip_code`` configuration value.
        :param units: Override the ``units`` configuration value.
        """
        params = {
            'units': units or self.units,
            **self._get_location_query(location=location, city_id=city_id, lat=lat, long=long)
        }

        response = self.get(self.base_url, params=params).output
        return self._convert_weather_response(response)



"""Flic client library for python

Requires python 3.3 or higher.

For detailed documentation, see the protocol documentation.

Notes on the data type used in this python implementation compared to the protocol documentation:
All kind of integers are represented as python integers.
Booleans use the Boolean type.
Enums use the defined python enums below.
Bd addr are represented as standard python strings, e.g. "aa:bb:cc:dd:ee:ff".
"""
import asyncio
from enum import Enum
from collections import namedtuple
import struct
import itertools
import threading


class CreateConnectionChannelError(Enum):
    NoError = 0
    MaxPendingConnectionsReached = 1


class ConnectionStatus(Enum):
    Disconnected = 0
    Connected = 1
    Ready = 2


class DisconnectReason(Enum):
    Unspecified = 0
    ConnectionEstablishmentFailed = 1
    TimedOut = 2
    BondingKeysMismatch = 3


class RemovedReason(Enum):
    RemovedByThisClient = 0
    ForceDisconnectedByThisClient = 1
    ForceDisconnectedByOtherClient = 2

    ButtonIsPrivate = 3
    VerifyTimeout = 4
    InternetBackendError = 5
    InvalidData = 6

    CouldntLoadDevice = 7


class ClickType(Enum):
    ButtonDown = 0
    ButtonUp = 1
    ButtonClick = 2
    ButtonSingleClick = 3
    ButtonDoubleClick = 4
    ButtonHold = 5


class BdAddrType(Enum):
    PublicBdAddrType = 0
    RandomBdAddrType = 1


class LatencyMode(Enum):
    NormalLatency = 0
    LowLatency = 1
    HighLatency = 2


class BluetoothControllerState(Enum):
    Detached = 0
    Resetting = 1
    Attached = 2


class ScanWizardResult(Enum):
    WizardSuccess = 0
    WizardCancelledByUser = 1
    WizardFailedTimeout = 2
    WizardButtonIsPrivate = 3
    WizardBluetoothUnavailable = 4
    WizardInternetBackendError = 5
    WizardInvalidData = 6


class ButtonScanner:
    """ButtonScanner class.

    Usage:
    scanner = ButtonScanner()
    scanner.on_advertisement_packet = lambda scanner, bd_addr, name, rssi, is_private, already_verified: ...
    client.add_scanner(scanner)
    """

    _cnt = itertools.count()

    def __init__(self):
        self._scan_id = next(ButtonScanner._cnt)
        self.on_advertisement_packet = lambda scanner, bd_addr, name, rssi, is_private, already_verified: None


class ScanWizard:
    """ScanWizard class

    Usage:
    wizard = ScanWizard()
    wizard.on_found_private_button = lambda scan_wizard: ...
    wizard.on_found_public_button = lambda scan_wizard, bd_addr, name: ...
    wizard.on_button_connected = lambda scan_wizard, bd_addr, name: ...
    wizard.on_completed = lambda scan_wizard, result, bd_addr, name: ...
    client.add_scan_wizard(wizard)
    """

    _cnt = itertools.count()

    def __init__(self):
        self._scan_wizard_id = next(ScanWizard._cnt)
        self._bd_addr = None
        self._name = None
        self.on_found_private_button = lambda scan_wizard: None
        self.on_found_public_button = lambda scan_wizard, bd_addr, name: None
        self.on_button_connected = lambda scan_wizard, bd_addr, name: None
        self.on_completed = lambda scan_wizard, result, bd_addr, name: None


class ButtonConnectionChannel:
    """ButtonConnectionChannel class.

    This class represents a connection channel to a Flic button.
    Add this button connection channel to a FlicClient by executing client.add_connection_channel(connection_channel).
    You may only have this connection channel added to one FlicClient at a time.

    Before you add the connection channel to the client, you should set up your callback functions by assigning
    the corresponding properties to this object with a function. Each callback function has a channel parameter as the first one,
    referencing this object.

    Available properties and the function parameters are:
    on_create_connection_channel_response: channel, error, connection_status
    on_removed: channel, removed_reason
    on_connection_status_changed: channel, connection_status, disconnect_reason
    on_button_up_or_down / on_button_click_or_hold / on_button_single_or_double_click / on_button_single_or_double_click_or_hold: channel, click_type, was_queued, time_diff
    """

    _cnt = itertools.count()

    def __init__(self, bd_addr, latency_mode=LatencyMode.NormalLatency, auto_disconnect_time=511):
        self._conn_id = next(ButtonConnectionChannel._cnt)
        self._bd_addr = bd_addr
        self._latency_mode = latency_mode
        self._auto_disconnect_time = auto_disconnect_time
        self._client = None

        self.on_create_connection_channel_response = lambda channel, error, connection_status: None
        self.on_removed = lambda channel, removed_reason: None
        self.on_connection_status_changed = lambda channel, connection_status, disconnect_reason: None
        self.on_button_up_or_down = lambda channel, click_type, was_queued, time_diff: None
        self.on_button_click_or_hold = lambda channel, click_type, was_queued, time_diff: None
        self.on_button_single_or_double_click = lambda channel, click_type, was_queued, time_diff: None
        self.on_button_single_or_double_click_or_hold = lambda channel, click_type, was_queued, time_diff: None

    @property
    def bd_addr(self):
        return self._bd_addr

    @property
    def latency_mode(self):
        return self._latency_mode

    @latency_mode.setter
    def latency_mode(self, latency_mode):
        if self._client is None:
            self._latency_mode = latency_mode
            return

        self._latency_mode = latency_mode
        if not self._client._closed:
            self._client._send_command("CmdChangeModeParameters",
                                       {"conn_id": self._conn_id, "latency_mode": self._latency_mode,
                                        "auto_disconnect_time": self._auto_disconnect_time})

    @property
    def auto_disconnect_time(self):
        return self._auto_disconnect_time

    @auto_disconnect_time.setter
    def auto_disconnect_time(self, auto_disconnect_time):
        if self._client is None:
            self._auto_disconnect_time = auto_disconnect_time
            return

        self._auto_disconnect_time = auto_disconnect_time
        if not self._client._closed:
            self._client._send_command("CmdChangeModeParameters",
                                       {"conn_id": self._conn_id, "latency_mode": self._latency_mode,
                                        "auto_disconnect_time": self._auto_disconnect_time})


class FlicClient(asyncio.Protocol):
    """FlicClient class.

    When this class is constructed, a socket connection is established.
    You may then send commands to the server and set timers.
    Once you are ready with the initialization you must call the handle_events() method which is a main loop that never exits, unless the socket is closed.
    For a more detailed description of all commands, events and enums, check the protocol specification.

    All commands are wrapped in more high level functions and events are reported using callback functions.

    All methods called on this class will take effect only if you eventually call the handle_events() method.

    The ButtonScanner is used to set up a handler for advertisement packets.
    The ButtonConnectionChannel is used to interact with connections to flic buttons and receive their events.

    Other events are handled by the following callback functions that can be assigned to this object (and a list of the callback function parameters):
    on_new_verified_button: bd_addr
    on_no_space_for_new_connection: max_concurrently_connected_buttons
    on_got_space_for_new_connection: max_concurrently_connected_buttons
    on_bluetooth_controller_state_change: state
    """

    _EVENTS = [
        ("EvtAdvertisementPacket", "<I6s17pb??", "scan_id bd_addr name rssi is_private already_verified"),
        ("EvtCreateConnectionChannelResponse", "<IBB", "conn_id error connection_status"),
        ("EvtConnectionStatusChanged", "<IBB", "conn_id connection_status disconnect_reason"),
        ("EvtConnectionChannelRemoved", "<IB", "conn_id removed_reason"),
        ("EvtButtonUpOrDown", "<IBBI", "conn_id click_type was_queued time_diff"),
        ("EvtButtonClickOrHold", "<IBBI", "conn_id click_type was_queued time_diff"),
        ("EvtButtonSingleOrDoubleClick", "<IBBI", "conn_id click_type was_queued time_diff"),
        ("EvtButtonSingleOrDoubleClickOrHold", "<IBBI", "conn_id click_type was_queued time_diff"),
        ("EvtNewVerifiedButton", "<6s", "bd_addr"),
        ("EvtGetInfoResponse", "<B6sBBhBBH",
         "bluetooth_controller_state my_bd_addr my_bd_addr_type max_pending_connections max_concurrently_connected_buttons current_pending_connections currently_no_space_for_new_connection nb_verified_buttons"),
        ("EvtNoSpaceForNewConnection", "<B", "max_concurrently_connected_buttons"),
        ("EvtGotSpaceForNewConnection", "<B", "max_concurrently_connected_buttons"),
        ("EvtBluetoothControllerStateChange", "<B", "state"),
        ("EvtPingResponse", "<I", "ping_id"),
        ("EvtGetButtonUUIDResponse", "<6s16s", "bd_addr uuid"),
        ("EvtScanWizardFoundPrivateButton", "<I", "scan_wizard_id"),
        ("EvtScanWizardFoundPublicButton", "<I6s17p", "scan_wizard_id bd_addr name"),
        ("EvtScanWizardButtonConnected", "<I", "scan_wizard_id"),
        ("EvtScanWizardCompleted", "<IB", "scan_wizard_id result")
    ]
    _EVENT_STRUCTS = list(map(lambda x: None if x is None else struct.Struct(x[1]), _EVENTS))
    _EVENT_NAMED_TUPLES = list(map(lambda x: None if x is None else namedtuple(x[0], x[2]), _EVENTS))

    _COMMANDS = [
        ("CmdGetInfo", "", ""),
        ("CmdCreateScanner", "<I", "scan_id"),
        ("CmdRemoveScanner", "<I", "scan_id"),
        ("CmdCreateConnectionChannel", "<I6sBh", "conn_id bd_addr latency_mode auto_disconnect_time"),
        ("CmdRemoveConnectionChannel", "<I", "conn_id"),
        ("CmdForceDisconnect", "<6s", "bd_addr"),
        ("CmdChangeModeParameters", "<IBh", "conn_id latency_mode auto_disconnect_time"),
        ("CmdPing", "<I", "ping_id"),
        ("CmdGetButtonUUID", "<6s", "bd_addr"),
        ("CmdCreateScanWizard", "<I", "scan_wizard_id"),
        ("CmdCancelScanWizard", "<I", "scan_wizard_id")
    ]

    _COMMAND_STRUCTS = list(map(lambda x: struct.Struct(x[1]), _COMMANDS))
    _COMMAND_NAMED_TUPLES = list(map(lambda x: namedtuple(x[0], x[2]), _COMMANDS))
    _COMMAND_NAME_TO_OPCODE = dict((x[0], i) for i, x in enumerate(_COMMANDS))

    @staticmethod
    def _bdaddr_bytes_to_string(bdaddr_bytes):
        return ":".join(map(lambda x: "%02x" % x, reversed(bdaddr_bytes)))

    @staticmethod
    def _bdaddr_string_to_bytes(bdaddr_string):
        return bytearray.fromhex("".join(reversed(bdaddr_string.split(":"))))

    def __init__(self, loop, parent=None):
        self.loop = loop
        self.buffer = b""
        self.transport = None
        self.parent = parent
        self._scanners = {}
        self._scan_wizards = {}
        self._connection_channels = {}
        self._closed = False

        self.on_new_verified_button = lambda bd_addr: None
        self.on_no_space_for_new_connection = lambda max_concurrently_connected_buttons: None
        self.on_got_space_for_new_connection = lambda max_concurrently_connected_buttons: None
        self.on_bluetooth_controller_state_change = lambda state: None
        self.on_get_info = lambda items: None
        self.on_get_button_uuid = lambda addr, uuid: None

    def connection_made(self, transport):
        self.transport = transport
        if self.parent:
            self.parent.register_protocol(self)

    def close(self):
        """Closes the client. The handle_events() method will return."""
        if self._closed:
            return

        self._closed = True

    def add_scanner(self, scanner):
        """Add a ButtonScanner object.

        The scan will start directly once the scanner is added.
        """
        if scanner._scan_id in self._scanners:
            return

        self._scanners[scanner._scan_id] = scanner
        self._send_command("CmdCreateScanner", {"scan_id": scanner._scan_id})

    def remove_scanner(self, scanner):
        """Remove a ButtonScanner object.

        You will no longer receive advertisement packets.
        """
        if scanner._scan_id not in self._scanners:
            return

        del self._scanners[scanner._scan_id]
        self._send_command("CmdRemoveScanner", {"scan_id": scanner._scan_id})

    def add_scan_wizard(self, scan_wizard):
        """Add a ScanWizard object.

        The scan wizard will start directly once the scan wizard is added.
        """
        if scan_wizard._scan_wizard_id in self._scan_wizards:
            return

        self._scan_wizards[scan_wizard._scan_wizard_id] = scan_wizard
        self._send_command("CmdCreateScanWizard", {"scan_wizard_id": scan_wizard._scan_wizard_id})

    def cancel_scan_wizard(self, scan_wizard):
        """Cancel a ScanWizard.

        Note: The effect of this command will take place at the time the on_completed event arrives on the scan wizard object.
        If cancelled due to this command, "result" in the on_completed event will be "WizardCancelledByUser".
        """
        if scan_wizard._scan_wizard_id not in self._scan_wizards:
            return

        self._send_command("CmdCancelScanWizard", {"scan_wizard_id": scan_wizard._scan_wizard_id})

    def add_connection_channel(self, channel):
        """Adds a connection channel to a specific Flic button.

        This will start listening for a specific Flic button's connection and button events.
        Make sure the Flic is either in public mode (by holding it down for 7 seconds) or already verified before calling this method.

        The on_create_connection_channel_response callback property will be called on the
        connection channel after this command has been received by the server.

        You may have as many connection channels as you wish for a specific Flic Button.
        """
        if channel._conn_id in self._connection_channels:
            return

        channel._client = self

        self._connection_channels[channel._conn_id] = channel
        self._send_command("CmdCreateConnectionChannel", {"conn_id": channel._conn_id, "bd_addr": channel.bd_addr,
                                                          "latency_mode": channel._latency_mode,
                                                          "auto_disconnect_time": channel._auto_disconnect_time})

    def remove_connection_channel(self, channel):
        """Remove a connection channel.

        This will stop listening for new events for a specific connection channel that has previously been added.
        Note: The effect of this command will take place at the time the on_removed event arrives on the connection channel object.
        """
        if channel._conn_id not in self._connection_channels:
            return

        self._send_command("CmdRemoveConnectionChannel", {"conn_id": channel._conn_id})

    def force_disconnect(self, bd_addr):
        """Force disconnection or cancel pending connection of a specific Flic button.

        This removes all connection channels for all clients connected to the server for this specific Flic button.
        """
        self._send_command("CmdForceDisconnect", {"bd_addr": bd_addr})

    def get_info(self):
        """Get info about the current state of the server.

        The server will send back its information directly and the callback will be called once the response arrives.
        The callback takes only one parameter: info. This info parameter is a dictionary with the following objects:
        bluetooth_controller_state, my_bd_addr, my_bd_addr_type, max_pending_connections, max_concurrently_connected_buttons,
        current_pending_connections, currently_no_space_for_new_connection, bd_addr_of_verified_buttons (a list of bd addresses).
        """
        self._send_command("CmdGetInfo", {})

    def get_button_uuid(self, bd_addr):
        """Get button uuid for a verified button.

        The server will send back its information directly and the callback will be called once the response arrives.
        Responses will arrive in the same order as requested.

        The callback takes two parameters: bd_addr, uuid (hex string of 32 characters).

        Note: if the button isn't verified, the uuid sent to the callback will rather be None.
        """
        self._send_command("CmdGetButtonUUID", {"bd_addr": bd_addr})

    def run_on_handle_events_thread(self, callback):
        """Run a function on the thread that handles the events."""
        if threading.get_ident() == self._handle_event_thread_ident:
            callback()
        else:
            self.set_timer(0, callback)

    def _send_command(self, name, items):

        for key, value in items.items():
            if isinstance(value, Enum):
                items[key] = value.value

        if "bd_addr" in items:
            items["bd_addr"] = FlicClient._bdaddr_string_to_bytes()

        opcode = FlicClient._COMMAND_NAME_TO_OPCODE[name]
        data_bytes = FlicClient._COMMAND_STRUCTS[opcode].pack(*FlicClient._COMMAND_NAMED_TUPLES[opcode](**items))
        bytes = bytearray(3)
        bytes[0] = (len(data_bytes) + 1) & 0xff
        bytes[1] = (len(data_bytes) + 1) >> 8
        bytes[2] = opcode
        bytes += data_bytes
        self.transport.write(bytes)

    def _dispatch_event(self, data):
        if len(data) == 0:
            return
        opcode = data[0]

        if opcode >= len(FlicClient._EVENTS) or FlicClient._EVENTS[opcode] is None:
            return

        event_name = FlicClient._EVENTS[opcode][0]
        data_tuple = FlicClient._EVENT_STRUCTS[opcode].unpack(data[1: 1 + FlicClient._EVENT_STRUCTS[opcode].size])
        items = FlicClient._EVENT_NAMED_TUPLES[opcode]._make(data_tuple)._asdict()

        # Process some kind of items whose data type is not supported by struct
        if "bd_addr" in items:
            items["bd_addr"] = FlicClient._bdaddr_bytes_to_string()

        if "name" in items:
            items["name"] = items["name"].decode("utf-8")

        if event_name == "EvtCreateConnectionChannelResponse":
            items["error"] = CreateConnectionChannelError(items["error"])
            items["connection_status"] = ConnectionStatus(items["connection_status"])

        if event_name == "EvtConnectionStatusChanged":
            items["connection_status"] = ConnectionStatus(items["connection_status"])
            items["disconnect_reason"] = DisconnectReason(items["disconnect_reason"])

        if event_name == "EvtConnectionChannelRemoved":
            items["removed_reason"] = RemovedReason(items["removed_reason"])

        if event_name.startswith("EvtButton"):
            items["click_type"] = ClickType(items["click_type"])

        if event_name == "EvtGetInfoResponse":
            items["bluetooth_controller_state"] = BluetoothControllerState(items["bluetooth_controller_state"])
            items["my_bd_addr"] = FlicClient._bdaddr_bytes_to_string()
            items["my_bd_addr_type"] = BdAddrType(items["my_bd_addr_type"])
            items["bd_addr_of_verified_buttons"] = []

            pos = FlicClient._EVENT_STRUCTS[opcode].size
            for i in range(items["nb_verified_buttons"]):
                items["bd_addr_of_verified_buttons"].append(
                    FlicClient._bdaddr_bytes_to_string())
                pos += 6

        if event_name == "EvtBluetoothControllerStateChange":
            items["state"] = BluetoothControllerState(items["state"])

        if event_name == "EvtGetButtonUUIDResponse":
            items["uuid"] = "".join(map(lambda x: "%02x" % x, items["uuid"]))
            if items["uuid"] == "00000000000000000000000000000000":
                items["uuid"] = None

        if event_name == "EvtScanWizardCompleted":
            items["result"] = ScanWizardResult(items["result"])

        # Process event
        if event_name == "EvtAdvertisementPacket":
            scanner = self._scanners.get(items["scan_id"])
            if scanner is not None:
                scanner.on_advertisement_packet(scanner, items["bd_addr"], items["name"], items["rssi"],
                                                items["is_private"], items["already_verified"])

        if event_name == "EvtCreateConnectionChannelResponse":
            channel = self._connection_channels[items["conn_id"]]
            if items["error"] != CreateConnectionChannelError.NoError:
                del self._connection_channels[items["conn_id"]]
            channel.on_create_connection_channel_response(channel, items["error"], items["connection_status"])

        if event_name == "EvtConnectionStatusChanged":
            channel = self._connection_channels[items["conn_id"]]
            channel.on_connection_status_changed(channel, items["connection_status"], items["disconnect_reason"])

        if event_name == "EvtConnectionChannelRemoved":
            channel = self._connection_channels[items["conn_id"]]
            del self._connection_channels[items["conn_id"]]
            channel.on_removed(channel, items["removed_reason"])

        if event_name == "EvtButtonUpOrDown":
            channel = self._connection_channels[items["conn_id"]]
            channel.on_button_up_or_down(channel, items["click_type"], items["was_queued"], items["time_diff"])
        if event_name == "EvtButtonClickOrHold":
            channel = self._connection_channels[items["conn_id"]]
            channel.on_button_click_or_hold(channel, items["click_type"], items["was_queued"], items["time_diff"])
        if event_name == "EvtButtonSingleOrDoubleClick":
            channel = self._connection_channels[items["conn_id"]]
            channel.on_button_single_or_double_click(channel, items["click_type"], items["was_queued"],
                                                     items["time_diff"])
        if event_name == "EvtButtonSingleOrDoubleClickOrHold":
            channel = self._connection_channels[items["conn_id"]]
            channel.on_button_single_or_double_click_or_hold(channel, items["click_type"], items["was_queued"],
                                                             items["time_diff"])

        if event_name == "EvtNewVerifiedButton":
            self.on_new_verified_button(items["bd_addr"])

        if event_name == "EvtGetInfoResponse":
            self.on_get_info(items)

        if event_name == "EvtNoSpaceForNewConnection":
            self.on_no_space_for_new_connection(items["max_concurrently_connected_buttons"])

        if event_name == "EvtGotSpaceForNewConnection":
            self.on_got_space_for_new_connection(items["max_concurrently_connected_buttons"])

        if event_name == "EvtBluetoothControllerStateChange":
            self.on_bluetooth_controller_state_change(items["state"])

        if event_name == "EvtGetButtonUUIDResponse":
            self.on_get_button_uuid(items["bd_addr"], items["uuid"])

        if event_name == "EvtScanWizardFoundPrivateButton":
            scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
            scan_wizard.on_found_private_button(scan_wizard)

        if event_name == "EvtScanWizardFoundPublicButton":
            scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
            scan_wizard._bd_addr = items["bd_addr"]
            scan_wizard._name = items["name"]
            scan_wizard.on_found_public_button(scan_wizard, scan_wizard._bd_addr, scan_wizard._name)

        if event_name == "EvtScanWizardButtonConnected":
            scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
            scan_wizard.on_button_connected(scan_wizard, scan_wizard._bd_addr, scan_wizard._name)

        if event_name == "EvtScanWizardCompleted":
            scan_wizard = self._scan_wizards[items["scan_wizard_id"]]
            del self._scan_wizards[items["scan_wizard_id"]]
            scan_wizard.on_completed(scan_wizard, items["result"], scan_wizard._bd_addr, scan_wizard._name)

    def data_received(self, data):
        cdata = self.buffer + data
        self.buffer = b""
        while len(cdata):
            packet_len = cdata[0] | (cdata[1] << 8)
            packet_len += 2
            if len(cdata) >= packet_len:
                self._dispatch_event(cdata[2:packet_len])
                cdata = cdata[packet_len:]
            else:
                if len(cdata):
                    self.buffer = cdata  # unlikely to happen but.....
                break



import copy
import json
import logging
import threading
from functools import wraps

from platypush.common import exec_wrapper
from platypush.config import Config
from platypush.message.event import Event
from platypush.message.request import Request
from platypush.procedure import Procedure
from platypush.utils import get_event_class_by_type, set_thread_name, is_functional_hook

logger = logging.getLogger('platypush')


def parse(msg):
    """Builds a dict given another dictionary or
    a JSON UTF-8 encoded string/bytearray"""

    if isinstance(msg, (bytes, bytearray)):
        msg = msg.decode('utf-8')
    if isinstance(msg, str):
        try:
            msg = json.loads(msg.strip())
        except json.JSONDecodeError:
            logger.warning('Invalid JSON message: {}'.format(msg))
            return None

    return msg


class EventCondition:
    """Event hook condition class"""

    def __init__(self, type=Event.__class__, priority=None, **kwargs):
        """
        Rule constructor.
        Params:
            type   -- Class of the event to be built
            kwargs -- Fields rules as a key-value (e.g. source_button=btn_id
                or recognized_phrase='Your phrase')
        """

        self.type = type
        self.args = {}
        self.parsed_args = {}
        self.priority = priority

        for (key, value) in kwargs.items():
            # TODO So far we only allow simple value match. If value is a dict
            # instead, we should allow more a sophisticated attribute matching,
            # e.g. or conditions, in, and other operators.
            self.args[key] = value

    @classmethod
    def build(cls, rule):
        """Builds a rule given either another EventRule, a dictionary or
        a JSON UTF-8 encoded string/bytearray"""

        if isinstance(rule, cls):
            return rule
        else:
            rule = parse(rule)

        assert isinstance(rule, dict), f'Not a valid rule: {rule}'
        type = get_event_class_by_type(rule.pop('type') if 'type' in rule else 'Event')

        args = {}
        for (key, value) in rule.items():
            args[key] = value

        return cls(type=type, **args)


class EventAction(Request):
    """Event hook action class. It is a special type of runnable request
    whose fields can be configured later depending on the event context"""

    def __init__(self, target=None, action=None, **args):
        if target is None:
            target = Config.get('device_id')
        args_copy = dict(copy.deepcopy(args))
        super().__init__(target=target, action=action, **args_copy)

    @classmethod
    def build(cls, action):
        action = super().parse(action)
        action['origin'] = Config.get('device_id')

        if 'target' not in action:
            action['target'] = action['origin']

        token = Config.get('token')
        if token:
            action['token'] = token

        return super().build(action)


class EventHook:
    """Event hook class. It consists of one conditions and
    one or multiple actions to be executed"""

    def __init__(self, name, priority=None, condition=None, actions=None):
        """Constructor. Takes a name, a EventCondition object and an event action
        procedure as input. It may also have a priority attached
        as a positive number. If multiple hooks match against an event,
        only the ones that have either the maximum match score or the
        maximum pre-configured priority will be run."""

        self.name = name
        self.condition = EventCondition.build(condition or {})
        self.actions = actions or []
        self.priority = priority or 0
        self.condition.priority = self.priority

    @classmethod
    def build(cls, name, hook):
        """Builds a rule given either another EventRule, a dictionary or
        a JSON UTF-8 encoded string/bytearray"""

        if isinstance(hook, cls):
            return hook
        else:
            hook = parse(hook)

        if is_functional_hook(hook):
            actions = Procedure(name=name, requests=[hook], _async=False)
            return cls(name=name, condition=hook.condition, actions=actions)

        assert isinstance(hook, dict)
        condition = EventCondition.build(hook['if']) if 'if' in hook else None
        actions = []
        priority = hook['priority'] if 'priority' in hook else None
        condition.priority = priority

        if 'then' in hook:
            if isinstance(hook['then'], list):
                actions = hook['then']
            else:
                actions = [hook['then']]

        actions = Procedure.build(name=name + '__Hook', requests=actions, _async=False)
        return cls(name=name, condition=condition, actions=actions, priority=priority)

    def matches_event(self, event):
        """Returns an EventMatchResult object containing the information
        about the match between the event and this hook"""

        return event.matches_condition(self.condition)

    def run(self, event):
        """Checks the condition of the hook against a particular event and
        runs the hook actions if the condition is met"""

        def _thread_func(result):
            set_thread_name('Event-' + self.name)
            self.actions.execute(event=event, **result.parsed_args)

        result = self.matches_event(event)

        if result.is_match:
            logger.info('Running hook {} triggered by an event'.format(self.name))
            threading.Thread(
                target=_thread_func, name='Event-' + self.name, args=(result,)
            ).start()


def hook(event_type=Event, **condition):
    def wrapper(f):
        f.hook = True
        f.condition = EventCondition(type=event_type, **condition)

        @wraps(f)
        def wrapped(event, *args, **kwargs):
            from platypush.message.event.http.hook import WebhookEvent

            response = exec_wrapper(f, event, *args, **kwargs)
            if isinstance(event, WebhookEvent):
                event.send_response(response)

            return response

        return wrapped

    return wrapper


# vim:sw=4:ts=4:et:



import json
import os
import pathlib

from dataclasses import dataclass
from datetime import datetime
from queue import Queue, Empty
from threading import Thread, RLock
from typing import List, Dict, Any, Optional, Tuple

from sqlalchemy import create_engine, Column, Integer, String, DateTime
import sqlalchemy.engine as engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base

from platypush.backend import Backend
from platypush.config import Config
from platypush.context import get_plugin
from platypush.message.event.mail import MailReceivedEvent, MailSeenEvent, MailFlaggedEvent, MailUnflaggedEvent
from platypush.plugins.mail import MailInPlugin, Mail

# <editor-fold desc="Database tables">
Base = declarative_base()
Session = scoped_session(sessionmaker())


class MailboxStatus(Base):
    """ Models the MailboxStatus table, containing information about the state of a monitored mailbox. """
    __tablename__ = 'MailboxStatus'

    mailbox_id = Column(Integer, primary_key=True)
    unseen_message_ids = Column(String, default='[]')
    flagged_message_ids = Column(String, default='[]')
    last_checked_date = Column(DateTime)


# </editor-fold>

# <editor-fold desc="Mailbox model">
@dataclass
class Mailbox:
    plugin: MailInPlugin
    name: str
    args: dict


# </editor-fold>


class MailBackend(Backend):
    """
    This backend can subscribe to one or multiple mail servers and trigger events when new messages are received or
    messages are marked as seen.

    It requires at least one plugin that extends :class:`platypush.plugins.mail.MailInPlugin` (e.g. ``mail.imap``) to
    be installed.

    Triggers:

        - :class:`platypush.message.event.mail.MailReceivedEvent` when a new message is received.
        - :class:`platypush.message.event.mail.MailSeenEvent` when a message is marked as seen.
        - :class:`platypush.message.event.mail.MailFlaggedEvent` when a message is marked as flagged/starred.
        - :class:`platypush.message.event.mail.MailUnflaggedEvent` when a message is marked as unflagged/unstarred.

    """

    def __init__(self, mailboxes: List[Dict[str, Any]], timeout: Optional[int] = 60, poll_seconds: Optional[int] = 60,
                 **kwargs):
        """
        :param mailboxes: List of mailboxes to be monitored. Each mailbox entry contains a ``plugin`` attribute to
            identify the :class:`platypush.plugins.mail.MailInPlugin` plugin that will be used (e.g. ``mail.imap``)
            and the arguments that will be passed to :meth:`platypush.plugins.mail.MailInPlugin.search_unseen_messages`.
            The ``name`` parameter can be used to identify this mailbox in the relevant events, otherwise
            ``Mailbox #{id}`` will be used as a name. Example configuration:

              .. code-block:: yaml

                 backend.mail:
                     mailboxes:
                         - plugin: mail.imap
                           name: "My Local Server"
                           username: me@mydomain.com
                           password: my-imap-password
                           server: localhost
                           ssl: true
                           folder: "All Mail"

                         - plugin: mail.imap
                           name: "GMail"
                           username: me@gmail.com
                           password: my-google-password
                           server: imap.gmail.com
                           ssl: true
                            folder: "INBOX"

            If you have a default configuration available for a mail plugin you can implicitly reuse it without
            replicating it here. Example:

              .. code-block:: yaml

                 mail.imap:
                     username: me@mydomain.com
                     password: my-imap-password
                     server: localhost
                     ssl: true

                 backend.mail:
                     mailboxes:
                         # The mail.imap default configuration will be used
                         - plugin: mail.imap
                           name: "My Local Server"
                           folder: "All Mail"

        :param poll_seconds: How often the backend should check the mail (default: 60).
        :param timeout: Connect/read timeout for a mailbox, in seconds (default: 60).
        """
        self.logger.info('Initializing mail backend')

        super().__init__(**kwargs)
        self.poll_seconds = poll_seconds
        self.mailboxes: List[Mailbox] = []
        self.timeout = timeout
        self._unread_msgs: List[Dict[int, Mail]] = [{}] * len(mailboxes)
        self._flagged_msgs: List[Dict[int, Mail]] = [{}] * len(mailboxes)
        self._db_lock = RLock()
        self.workdir = os.path.join(os.path.expanduser(Config.get('workdir')), 'mail')
        self.dbfile = os.path.join(self.workdir, 'backend.db')

        # Parse mailboxes
        for i, mbox in enumerate(mailboxes):
            assert 'plugin' in mbox, 'No plugin attribute specified for mailbox n.{}'.format(i)
            plugin = get_plugin(mbox.pop('plugin'))
            assert isinstance(plugin, MailInPlugin), '{} is not a MailInPlugin'.format(plugin)
            name = mbox.pop('name') if 'name' in mbox else 'Mailbox #{}'.format(i + 1)
            self.mailboxes.append(Mailbox(plugin=plugin, name=name, args=mbox))

        # Configure/sync db
        pathlib.Path(self.workdir).mkdir(parents=True, exist_ok=True, mode=0o750)
        self._db = self._db_get_engine()
        Base.metadata.create_all(self._db)
        Session.configure(bind=self._db)
        self._db_load_mailboxes_status()
        self.logger.info('Mail backend initialized')

    # <editor-fold desc="Database methods">
    def _db_get_engine(self) -> engine.Engine:
        return create_engine('sqlite:///{}'.format(self.dbfile), connect_args={'check_same_thread': False})

    def _db_load_mailboxes_status(self) -> None:
        mailbox_ids = list(range(len(self.mailboxes)))

        with self._db_lock:
            session = Session()
            records = {
                record.mailbox_id: record
                for record in session.query(MailboxStatus).filter(MailboxStatus.mailbox_id.in_(mailbox_ids)).all()
            }

            for mbox_id, mbox in enumerate(self.mailboxes):
                if mbox_id not in records:
                    record = MailboxStatus(mailbox_id=mbox_id, unseen_message_ids='[]', flagged_message_ids='[]')
                    session.add(record)
                else:
                    record = records[mbox_id]

                unseen_msg_ids = json.loads(record.unseen_message_ids or '[]')
                flagged_msg_ids = json.loads(record.flagged_message_ids or '[]')
                self._unread_msgs[mbox_id] = {msg_id: {} for msg_id in unseen_msg_ids}
                self._flagged_msgs[mbox_id] = {msg_id: {} for msg_id in flagged_msg_ids}

            session.commit()

    def _db_get_mailbox_status(self, mailbox_ids: List[int]) -> Dict[int, MailboxStatus]:
        with self._db_lock:
            session = Session()
            return {
                record.mailbox_id: record
                for record in session.query(MailboxStatus).filter(MailboxStatus.mailbox_id.in_(mailbox_ids)).all()
            }

    # </editor-fold>

    # <editor-fold desc="Parse unread messages logic">
    @staticmethod
    def _check_thread(unread_queue: Queue, flagged_queue: Queue, plugin: MailInPlugin, **args):
        def thread():
            # noinspection PyUnresolvedReferences
            unread = plugin.search_unseen_messages(**args).output
            unread_queue.put({msg.id: msg for msg in unread})

            # noinspection PyUnresolvedReferences
            flagged = plugin.search_flagged_messages(**args).output
            flagged_queue.put({msg.id: msg for msg in flagged})

        return thread

    def _get_unread_seen_msgs(self, mailbox_idx: int, unread_msgs: Dict[int, Mail]) \
            -> Tuple[Dict[int, Mail], Dict[int, Mail]]:
        prev_unread_msgs = self._unread_msgs[mailbox_idx]

        return {
            msg_id: unread_msgs[msg_id]
            for msg_id in unread_msgs
            if msg_id not in prev_unread_msgs
        }, {
            msg_id: prev_unread_msgs[msg_id]
            for msg_id in prev_unread_msgs
            if msg_id not in unread_msgs
        }

    def _get_flagged_unflagged_msgs(self, mailbox_idx: int, flagged_msgs: Dict[int, Mail]) \
            -> Tuple[Dict[int, Mail], Dict[int, Mail]]:
        prev_flagged_msgs = self._flagged_msgs[mailbox_idx]

        return {
                   msg_id: flagged_msgs[msg_id]
                   for msg_id in flagged_msgs
                   if msg_id not in prev_flagged_msgs
               }, {
                   msg_id: prev_flagged_msgs[msg_id]
                   for msg_id in prev_flagged_msgs
                   if msg_id not in flagged_msgs
               }

    def _process_msg_events(self, mailbox_id: int, unread: List[Mail], seen: List[Mail],
                            flagged: List[Mail], unflagged: List[Mail], last_checked_date: Optional[datetime] = None):
        for msg in unread:
            if msg.date and last_checked_date and msg.date < last_checked_date:
                continue
            self.bus.post(MailReceivedEvent(mailbox=self.mailboxes[mailbox_id].name, message=msg))

        for msg in seen:
            self.bus.post(MailSeenEvent(mailbox=self.mailboxes[mailbox_id].name, message=msg))

        for msg in flagged:
            self.bus.post(MailFlaggedEvent(mailbox=self.mailboxes[mailbox_id].name, message=msg))

        for msg in unflagged:
            self.bus.post(MailUnflaggedEvent(mailbox=self.mailboxes[mailbox_id].name, message=msg))

    def _check_mailboxes(self) -> List[Tuple[Dict[int, Mail], Dict[int, Mail]]]:
        workers = []
        queues: List[Tuple[Queue, Queue]] = []
        results = []

        for mbox in self.mailboxes:
            unread_queue, flagged_queue = [Queue()] * 2
            worker = Thread(target=self._check_thread(unread_queue=unread_queue, flagged_queue=flagged_queue,
                                                      plugin=mbox.plugin, **mbox.args))
            worker.start()
            workers.append(worker)
            queues.append((unread_queue, flagged_queue))

        for worker in workers:
            worker.join(timeout=self.timeout)

        for i, (unread_queue, flagged_queue) in enumerate(queues):
            try:
                unread = unread_queue.get(timeout=self.timeout)
                flagged = flagged_queue.get(timeout=self.timeout)
                results.append((unread, flagged))
            except Empty:
                self.logger.warning('Checks on mailbox #{} timed out after {} seconds'.format(i + 1, self.timeout))
                continue

        return results

    # </editor-fold>

    # <editor-fold desc="Loop function">
    def loop(self):
        records = []
        mailbox_statuses = self._db_get_mailbox_status(list(range(len(self.mailboxes))))
        results = self._check_mailboxes()

        for i, (unread, flagged) in enumerate(results):
            unread_msgs, seen_msgs = self._get_unread_seen_msgs(i, unread)
            flagged_msgs, unflagged_msgs = self._get_flagged_unflagged_msgs(i, flagged)
            self._process_msg_events(i, unread=list(unread_msgs.values()), seen=list(seen_msgs.values()),
                                     flagged=list(flagged_msgs.values()), unflagged=list(unflagged_msgs.values()),
                                     last_checked_date=mailbox_statuses[i].last_checked_date)

            self._unread_msgs[i] = unread
            self._flagged_msgs[i] = flagged
            records.append(MailboxStatus(mailbox_id=i,
                                         unseen_message_ids=json.dumps([msg_id for msg_id in unread.keys()]),
                                         flagged_message_ids=json.dumps([msg_id for msg_id in flagged.keys()]),
                                         last_checked_date=datetime.now()))

        with self._db_lock:
            session = Session()
            for record in records:
                session.merge(record)
            session.commit()

    # </editor-fold>


# vim:sw=4:ts=4:et:



from typing import Iterable, Dict, Union, Any

from watchdog.observers import Observer

from platypush.backend import Backend
from .entities.handlers import EventHandler
from .entities.resources import MonitoredResource, MonitoredPattern, MonitoredRegex


class FileMonitorBackend(Backend):
    """
    This backend monitors changes to local files and directories using the Watchdog API.

    Triggers:

        * :class:`platypush.message.event.file.FileSystemCreateEvent` if a resource is created.
        * :class:`platypush.message.event.file.FileSystemDeleteEvent` if a resource is removed.
        * :class:`platypush.message.event.file.FileSystemModifyEvent` if a resource is modified.

    Requires:

        * **watchdog** (``pip install watchdog``)

    """

    class EventHandlerFactory:
        """
        Create a file system event handler from a string, dictionary or ``MonitoredResource`` resource.
        """

        @staticmethod
        def from_resource(resource: Union[str, Dict[str, Any], MonitoredResource]) -> EventHandler:
            if isinstance(resource, str):
                resource = MonitoredResource(resource)
            elif isinstance(resource, dict):
                if 'regexes' in resource or 'ignore_regexes' in resource:
                    resource = MonitoredRegex(**resource)
                elif 'patterns' in resource or 'ignore_patterns' in resource or 'ignore_directories' in resource:
                    resource = MonitoredPattern(**resource)
                else:
                    resource = MonitoredResource(**resource)

            return EventHandler.from_resource(resource)

    def __init__(self, paths: Iterable[Union[str, Dict[str, Any], MonitoredResource]], **kwargs):
        """
        :param paths: List of paths to monitor. Paths can either be expressed in any of the following ways:

            - Simple strings. In this case, paths will be interpreted as absolute references to a file or a directory
              to monitor. Example:

              .. code-block:: yaml

                backend.file.monitor:
                    paths:
                        # Monitor changes on the /tmp folder
                        - /tmp
                        # Monitor changes on /etc/passwd
                        - /etc/passwd

            - Path with monitoring properties expressed as a key-value object. Example showing the supported attributes:

                .. code-block:: yaml

                    backend.file.monitor:
                        paths:
                            # Monitor changes on the /tmp folder and its subfolders
                            - path: /tmp
                              recursive: True

            - Path with pattern-based search criteria for the files to monitor and exclude. Example:

                .. code-block:: yaml

                    backend.file.monitor:
                        paths:
                            # Recursively monitor changes on the ~/my-project folder that include all
                            # *.py files, excluding those whose name starts with tmp_ and
                            # all the files contained in the __pycache__ folders
                            - path: ~/my-project
                              recursive: True
                              patterns:
                                - "*.py"
                              ignore_patterns:
                                - "tmp_*"
                              ignore_directories:
                                - "__pycache__"

            - Path with regex-based search criteria for the files to monitor and exclude. Example:

                .. code-block:: yaml

                    backend.file.monitor:
                        paths:
                            # Recursively monitor changes on the ~/my-images folder that include all
                            # the files matching a JPEG extension in case-insensitive mode,
                            # excluding those whose name starts with tmp_ and
                            # all the files contained in the __MACOSX folders
                            - path: ~/my-images
                              recursive: True
                              case_sensitive: False
                              regexes:
                                - '.*\\.jpe?g$'
                              ignore_patterns:
                                - '^tmp_.*'
                              ignore_directories:
                                - '__MACOSX'

        """

        super().__init__(**kwargs)
        self._observer = Observer()

        for path in paths:
            handler = self.EventHandlerFactory.from_resource(path)
            self._observer.schedule(handler, handler.resource.path, recursive=handler.resource.recursive)

    def run(self):
        super().run()
        self.logger.info('Initializing file monitor backend')
        self._observer.start()
        self.wait_stop()

    def on_stop(self):
        self.logger.info('Stopping file monitor backend')
        self._observer.stop()
        self._observer.join()
        self.logger.info('Stopped file monitor backend')



import io
import logging
import os
import threading
import time

from abc import ABC, abstractmethod
from typing import Optional, IO

from PIL.Image import Image


class VideoWriter(ABC):
    """
    Generic class interface for handling frames-to-video operations.
    """

    mimetype: Optional[str] = None

    def __init__(self, camera, plugin, *_, **__):
        from platypush.plugins.camera import Camera, CameraPlugin

        self.logger = logging.getLogger(self.__class__.__name__)
        self.camera: Camera = camera
        self.plugin: CameraPlugin = plugin
        self.closed = False

    @abstractmethod
    def write(self, img: Image):
        """
        Write an image to the channel.

        :param img: PIL Image instance.
        """
        raise NotImplementedError()

    @abstractmethod
    def close(self):
        """
        Close the channel.
        """
        if self.camera:
            self.plugin.close_device(self.camera)
        self.closed = True

    def __enter__(self):
        """
        Context manager-based interface.
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Context manager-based interface.
        """
        self.close()


class FileVideoWriter(VideoWriter, ABC):
    """
    Abstract class to handle frames-to-video file operations.
    """
    def __init__(self, *args, output_file: str, **kwargs):
        VideoWriter.__init__(self, *args, **kwargs)
        self.output_file = os.path.abspath(os.path.expanduser(output_file))


class StreamWriter(VideoWriter, ABC):
    """
    Abstract class for camera streaming operations.
    """
    def __init__(self, *args, sock: Optional[IO] = None, **kwargs):
        VideoWriter.__init__(self, *args, **kwargs)
        self.frame: Optional[bytes] = None
        self.frame_time: Optional[float] = None
        self.buffer = io.BytesIO()
        self.ready = threading.Condition()
        self.sock = sock

    def write(self, image: Image):
        data = self.encode(image)
        with self.ready:
            if self.buffer.closed:
                return

            self.buffer.truncate()
            self.frame = self.buffer.getvalue()
            self.frame_time = time.time()
            self.ready.notify_all()

        self._sock_send(self.frame)
        if not self.buffer.closed:
            self.buffer.seek(0)
            return self.buffer.write(data)

    def _sock_send(self, data):
        if self.sock and data:
            try:
                self.sock.write(data)
            except ConnectionError:
                self.logger.info('Client connection closed')
                self.close()

    @abstractmethod
    def encode(self, image: Image) -> bytes:
        """
        Encode an image before sending it to the channel.

        :param image: PIL Image object.
        :return: The bytes-encoded representation of the frame.
        """
        raise NotImplementedError()

    def close(self):
        self.buffer.close()
        if self.sock:
            try:
                self.sock.close()
            except Exception as e:
                self.logger.warning('Could not close camera resource: {}'.format(str(e)))

        super().close()

    @staticmethod
    def get_class_by_name(name: str):
        from platypush.plugins.camera.model.writer.index import StreamHandlers
        name = name.upper()
        assert hasattr(StreamHandlers, name), 'No such stream handler: {}. Supported types: {}'.format(
            name, [hndl.name for hndl in list(StreamHandlers)])

        return getattr(StreamHandlers, name).value


# vim:sw=4:ts=4:et:



import datetime
import os
import re

from dataclasses import dataclass
from logging import getLogger
from threading import RLock
from typing import List, Optional, Iterable

from platypush.backend.file.monitor import FileMonitorBackend, EventHandler, MonitoredResource
from platypush.context import get_bus
from platypush.message.event.log.http import HttpLogEvent

logger = getLogger(__name__)


class LogEventHandler(EventHandler):
    http_line_regex = re.compile(r'^([^\s]+)\s+([^\s]+)\s+([^\s]+)\s+\[([^]]+)]\s+"([^"]+)"\s+([\d]+)\s+'
                                 r'([\d]+)\s*("([^"\s]+)")?\s*("([^"]+)")?$')

    @dataclass
    class FileResource:
        path: str
        pos: int = 0
        lock: RLock = RLock()
        last_timestamp: Optional[datetime.datetime] = None

    def __init__(self, *args, monitored_files: Optional[Iterable[str]] = None, **kwargs):
        super().__init__(*args, **kwargs)
        self._monitored_files = {}
        self.monitor_files(monitored_files or [])

    def monitor_files(self, files: Iterable[str]):
        self._monitored_files.update({
            f: self.FileResource(path=f, pos=self._get_size(f))
            for f in files
        })

    @staticmethod
    def _get_size(file: str) -> int:
        try:
            return os.path.getsize(file)
        except Exception as e:
            logger.warning('Could not open {}: {}'.format(file, str(e)))
            return 0

    def on_created(self, event):
        self._reset_file(event.src_path)

    def on_deleted(self, event):
        self._reset_file(event.src_path)

    def on_moved(self, event):
        self._reset_file(event.src_path)

    def _reset_file(self, path: str):
        file_info = self._monitored_files.get(path)
        if not file_info:
            return

        file_info.pos = 0

    def on_modified(self, event):
        file_info = self._monitored_files.get(event.src_path)
        if not file_info:
            return

        try:
            file_size = os.path.getsize(event.src_path)
        except OSError as e:
            logger.warning('Could not get the size of {}: {}'.format(event.src_path, str(e)))
            return

        if file_info.pos > file_size:
            logger.warning('The size of {} been unexpectedly decreased from {} to {} bytes'.format(
                event.src_path, file_info.pos, file_size))
            file_info.pos = 0

        try:
            with file_info.lock, open(event.src_path, 'r') as f:
                f.seek(file_info.pos)
                for line in f.readlines():
                    evt = self._build_event(file=event.src_path, line=line)
                    if evt and (not file_info.last_timestamp or evt.args['time'] >= file_info.last_timestamp):
                        get_bus().post(evt)
                        file_info.last_timestamp = evt.args['time']

                file_info.pos = f.tell()
        except OSError as e:
            logger.warning('Error while reading from {}: {}'.format(self.resource.path, str(e)))

    @classmethod
    def _build_event(cls, file: str, line: str) -> Optional[HttpLogEvent]:
        line = line.strip()
        if not line:
            return

        m = cls.http_line_regex.match(line.strip())
        if not m:
            logger.warning('Could not parse log line from {}: {}'.format(file, line))
            return

        url = None
        method = 'GET'
        http_version = '1.0'

        try:
            url = m.group(5).split(' ')[1]
            method = m.group(5).split(' ')[0]
            http_version = m.group(5).split(' ')[2].split('/')[1]
        except Exception as e:
            logger.debug(str(e))

        if not url:
            return

        info = {
            'address': m.group(1),
            'user_identifier': m.group(2),
            'user_id': m.group(3),
            'time': datetime.datetime.strptime(m.group(4), '%d/%b/%Y:%H:%M:%S %z'),
            'method': method,
            'url': url,
            'http_version': http_version,
            'status': int(m.group(6)),
            'size': int(m.group(7)),
            'referrer': m.group(9),
            'user_agent': m.group(11),
        }

        for attr, value in info.items():
            if value == '-':
                info[attr] = None

        return HttpLogEvent(logfile=file, **info)


class LogHttpBackend(FileMonitorBackend):
    """
    This backend can be used to monitor one or more HTTP log files (tested on Apache and Nginx) and trigger events
    whenever a new log line is added.

    Triggers:

        * :class:`platypush.message.event.log.http.HttpLogEvent` when a new log line is created.

    Requires:

        * **watchdog** (``pip install watchdog``)

    """

    class EventHandlerFactory:
        @staticmethod
        def from_resource(resource: str) -> LogEventHandler:
            resource = MonitoredResource(resource)
            return LogEventHandler.from_resource(resource)

    def __init__(self, log_files: List[str], **kwargs):
        """
        :param log_files: List of log files to be monitored.
        """
        self.log_files = {os.path.expanduser(log) for log in log_files}
        directories = {os.path.dirname(log) for log in self.log_files}
        super().__init__(paths=directories, **kwargs)

        # noinspection PyProtectedMember
        handlers = self._observer._handlers
        for hndls in handlers.values():
            for hndl in hndls:
                hndl.monitor_files(self.log_files)



from typing import Optional, Union

from platypush.plugins.camera import CameraPlugin, Camera
from platypush.plugins.camera.model.writer.cv import CvFileWriter


class CameraCvPlugin(CameraPlugin):
    """
    Plugin to control generic cameras over OpenCV.

    Requires:

        * **opencv** (``pip install opencv-python``)
        * **Pillow** (``pip install Pillow``)

    """

    def __init__(self, color_transform: Optional[str] = 'COLOR_BGR2RGB', video_type: str = 'XVID',
                 video_writer: str = 'ffmpeg', **kwargs):
        """
        :param device: Device ID (0 for the first camera, 1 for the second etc.) or path (e.g. ``/dev/video0``).
        :param video_type: Default video type to use when exporting captured frames to camera (default: 0, infers the
            type from the video file extension). See
            `here <https://docs.opencv.org/4.0.1/dd/d9e/classcv_1_1VideoWriter.html#afec93f94dc6c0b3e28f4dd153bc5a7f0>`_
            for a reference on the supported types (e.g. 'MJPEG', 'XVID', 'H264', 'X264', 'AVC1' etc.)

        :param color_transform: Color transformation to apply to the captured frames. See
            https://docs.opencv.org/3.2.0/d7/d1b/group__imgproc__misc.html for a full list of supported color
            transformations (default: "``COLOR_RGB2BGR``")

        :param video_writer: Class to be used to write frames to a video file. Supported values:

                - ``ffmpeg``: Use the FFmpeg writer (default, and usually more reliable - it requires ``ffmpeg``
                    installed).
                - ``cv``: Use the native OpenCV writer.

            The FFmpeg video writer requires ``scikit-video`` (``pip install scikit-video``) and ``ffmpeg``.

        :param kwargs: Extra arguments to be passed up to :class:`platypush.plugins.camera.CameraPlugin`.
        """
        super().__init__(color_transform=color_transform, video_type=video_type, **kwargs)
        if video_writer == 'cv':
            self._video_writer_class = CvFileWriter

    def prepare_device(self, device: Camera):
        import cv2

        cam = cv2.VideoCapture(device.info.device)
        if device.info.resolution and device.info.resolution[0]:
            cam.set(cv2.CAP_PROP_FRAME_WIDTH, device.info.resolution[0])
            cam.set(cv2.CAP_PROP_FRAME_HEIGHT, device.info.resolution[1])

        return cam

    def release_device(self, device: Camera):
        if device.object:
            device.object.release()
            device.object = None

    def capture_frame(self, camera: Camera, *args, **kwargs):
        import cv2
        from PIL import Image
        ret, frame = camera.object.read()
        assert ret, 'Cannot retrieve frame from {}'.format(camera.info.device)

        color_transform = camera.info.color_transform
        if isinstance(color_transform, str):
            color_transform = getattr(cv2, color_transform or self.camera_info.color_transform)
        if color_transform:
            frame = cv2.cvtColor(frame, color_transform)

        return Image.fromarray(frame)

    @staticmethod
    def transform_frame(frame, color_transform: Union[str, int]):
        return frame


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod

from platypush.plugins import Plugin, action


class WeatherPlugin(Plugin, ABC):
    """
    Base class for weather plugins.
    """

    @action
    @abstractmethod
    def get_current_weather(self, *args, **kwargs):
        raise NotImplementedError



import importlib
import inspect
import json
import pkgutil
import re
import threading
from abc import ABC, abstractmethod
from typing import Optional

import platypush.backend   # lgtm [py/import-and-import-from]
import platypush.plugins   # lgtm [py/import-and-import-from]
import platypush.message.event   # lgtm [py/import-and-import-from]
import platypush.message.response   # lgtm [py/import-and-import-from]

from platypush.backend import Backend
from platypush.config import Config
from platypush.plugins import Plugin, action
from platypush.message.event import Event
from platypush.message.response import Response
from platypush.utils import get_decorators


class Model(ABC):
    def __str__(self):
        return json.dumps(dict(self), indent=2, sort_keys=True)

    def __repr__(self):
        return json.dumps(dict(self))

    @staticmethod
    def to_html(doc):
        try:
            import docutils.core
        except ImportError:
            # docutils not found
            return doc

        return docutils.core.publish_parts(doc, writer_name='html')['html_body']

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()


class ProcedureEncoder(json.JSONEncoder):
    def default(self, o):
        if callable(o):
            return {
                'type': 'native_function',
            }

        return super().default(o)


class BackendModel(Model):
    def __init__(self, backend, prefix='', html_doc: Optional[bool] = False):
        self.name = backend.__module__[len(prefix):]
        self.html_doc = html_doc
        self.doc = self.to_html(backend.__doc__) if html_doc and backend.__doc__ else backend.__doc__

    def __iter__(self):
        for attr in ['name', 'doc', 'html_doc']:
            yield attr, getattr(self, attr)


class PluginModel(Model):
    def __init__(self, plugin, prefix='', html_doc: Optional[bool] = False):
        self.name = plugin.__module__[len(prefix):]
        self.html_doc = html_doc
        self.doc = self.to_html(plugin.__doc__) if html_doc and plugin.__doc__ else plugin.__doc__
        self.actions = {action_name: ActionModel(getattr(plugin, action_name), html_doc=html_doc or False)
                        for action_name in get_decorators(plugin, climb_class_hierarchy=True).get('action', [])}

    def __iter__(self):
        for attr in ['name', 'actions', 'doc', 'html_doc']:
            if attr == 'actions':
                # noinspection PyShadowingNames
                yield attr, {name: dict(action) for name, action in self.actions.items()},
            else:
                yield attr, getattr(self, attr)


class EventModel(Model):
    def __init__(self, event, prefix='', html_doc: Optional[bool] = False):
        self.package = event.__module__[len(prefix):]
        self.name = event.__name__
        self.html_doc = html_doc
        self.doc = self.to_html(event.__doc__) if html_doc and event.__doc__ else event.__doc__

    def __iter__(self):
        for attr in ['name', 'doc', 'html_doc']:
            yield attr, getattr(self, attr)


class ResponseModel(Model):
    def __init__(self, response, prefix='', html_doc: Optional[bool] = False):
        self.package = response.__module__[len(prefix):]
        self.name = response.__name__
        self.html_doc = html_doc
        self.doc = self.to_html(response.__doc__) if html_doc and response.__doc__ else response.__doc__

    def __iter__(self):
        for attr in ['name', 'doc', 'html_doc']:
            yield attr, getattr(self, attr)


class ActionModel(Model):
    # noinspection PyShadowingNames
    def __init__(self, action, html_doc: bool = False):
        self.name = action.__name__
        self.doc, argsdoc = self._parse_docstring(action.__doc__, html_doc=html_doc)
        self.args = {}
        self.has_kwargs = False

        for arg in list(inspect.signature(action).parameters.values())[1:]:
            if arg.kind == arg.VAR_KEYWORD:
                self.has_kwargs = True
                continue

            self.args[arg.name] = {
                'default': arg.default if not issubclass(arg.default.__class__, type) else None,
                'doc': argsdoc.get(arg.name)
            }

    @classmethod
    def _parse_docstring(cls, docstring: str, html_doc: bool = False):
        new_docstring = ''
        params = {}
        cur_param = None
        cur_param_docstring = ''

        if not docstring:
            return None, {}

        for line in docstring.split('\n'):
            m = re.match(r'^\s*:param ([^:]+):\s*(.*)', line)
            if m:
                if cur_param:
                    params[cur_param] = cls.to_html(cur_param_docstring) if html_doc else cur_param_docstring

                cur_param = m.group(1)
                cur_param_docstring = m.group(2)
            elif re.match(r'^\s*:[^:]+:\s*.*', line):
                continue
            else:
                if cur_param:
                    if not line.strip():
                        params[cur_param] = cls.to_html(cur_param_docstring) if html_doc else cur_param_docstring
                        cur_param = None
                        cur_param_docstring = ''
                    else:
                        cur_param_docstring += '\n' + line.strip()
                else:
                    new_docstring += line.rstrip() + '\n'

        if cur_param:
            params[cur_param] = cls.to_html(cur_param_docstring) if html_doc else cur_param_docstring

        return new_docstring.strip() if not html_doc else cls.to_html(new_docstring), params

    def __iter__(self):
        for attr in ['name', 'args', 'doc', 'has_kwargs']:
            yield attr, getattr(self, attr)


class InspectPlugin(Plugin):
    """
    This plugin can be used to inspect platypush plugins and backends

    Requires:

        * **docutils** (``pip install docutils``) - optional, for HTML doc generation

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._plugins = {}
        self._backends = {}
        self._events = {}
        self._responses = {}
        self._plugins_lock = threading.RLock()
        self._backends_lock = threading.RLock()
        self._events_lock = threading.RLock()
        self._responses_lock = threading.RLock()
        self._html_doc = False

    def _init_plugins(self):
        package = platypush.plugins
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda _: None):
            try:
                module = importlib.import_module(modname)
            except Exception as e:
                self.logger.warning(f'Could not import module {modname}')
                self.logger.exception(e)
                continue

            for _, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and issubclass(obj, Plugin):
                    model = PluginModel(plugin=obj, prefix=prefix, html_doc=self._html_doc)
                    if model.name:
                        self._plugins[model.name] = model

    def _init_backends(self):
        package = platypush.backend
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda _: None):
            try:
                module = importlib.import_module(modname)
            except Exception as e:
                self.logger.debug(f'Could not import module {modname}: {str(e)}')
                continue

            for _, obj in inspect.getmembers(module):
                if inspect.isclass(obj) and issubclass(obj, Backend):
                    model = BackendModel(backend=obj, prefix=prefix, html_doc=self._html_doc)
                    if model.name:
                        self._backends[model.name] = model

    def _init_events(self):
        package = platypush.message.event
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda _: None):
            try:
                module = importlib.import_module(modname)
            except Exception as e:
                self.logger.debug(f'Could not import module {modname}: {str(e)}')
                continue

            for _, obj in inspect.getmembers(module):
                if type(obj) == Event:
                    continue

                if inspect.isclass(obj) and issubclass(obj, Event) and obj != Event:
                    event = EventModel(event=obj, html_doc=self._html_doc, prefix=prefix)
                    if event.package not in self._events:
                        self._events[event.package] = {event.name: event}
                    else:
                        self._events[event.package][event.name] = event

    def _init_responses(self):
        package = platypush.message.response
        prefix = package.__name__ + '.'

        for _, modname, _ in pkgutil.walk_packages(path=package.__path__,
                                                   prefix=prefix,
                                                   onerror=lambda _: None):
            try:
                module = importlib.import_module(modname)
            except Exception as e:
                self.logger.debug(f'Could not import module {modname}: {str(e)}')
                continue

            for _, obj in inspect.getmembers(module):
                if type(obj) == Response:
                    continue

                if inspect.isclass(obj) and issubclass(obj, Response) and obj != Response:
                    response = ResponseModel(response=obj, html_doc=self._html_doc, prefix=prefix)
                    if response.package not in self._responses:
                        self._responses[response.package] = {response.name: response}
                    else:
                        self._responses[response.package][response.name] = response

    @action
    def get_all_plugins(self, html_doc: Optional[bool] = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._plugins_lock:
            if not self._plugins or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_plugins()

            return json.dumps({
                name: dict(plugin)
                for name, plugin in self._plugins.items()
            })

    @action
    def get_all_backends(self, html_doc: Optional[bool] = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._backends_lock:
            if not self._backends or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_backends()

            return json.dumps({
                name: dict(backend)
                for name, backend in self._backends.items()
            })

    @action
    def get_all_events(self, html_doc: Optional[bool] = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._events_lock:
            if not self._events or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_events()

            return json.dumps({
                package: {
                    name: dict(event)
                    for name, event in events.items()
                }
                for package, events in self._events.items()
            })

    @action
    def get_all_responses(self, html_doc: Optional[bool] = None):
        """
        :param html_doc: If True then the docstring will be parsed into HTML (default: False)
        """
        with self._responses_lock:
            if not self._responses or (html_doc is not None and html_doc != self._html_doc):
                self._html_doc = html_doc
                self._init_responses()

            return json.dumps({
                package: {
                    name: dict(event)
                    for name, event in responses.items()
                }
                for package, responses in self._responses.items()
            })

    @action
    def get_procedures(self) -> dict:
        """
        Get the list of procedures installed on the device.
        """
        return json.loads(json.dumps(Config.get_procedures(), cls=ProcedureEncoder))

    @action
    def get_config(self, entry: Optional[str] = None) -> Optional[dict]:
        """
        Return the configuration of the application or of a section.

        :param entry: [Optional] configuration entry name to retrieve (e.g. ``workdir`` or ``backend.http``).
        :return: The requested configuration object.
        """
        if entry:
            return Config.get(entry)

        cfg = Config.get()
        return cfg


# vim:sw=4:ts=4:et:



import math
import threading
from dataclasses import dataclass
from typing import Optional, Union, Tuple, Set

import numpy as np

from platypush.plugins.camera.model.writer import StreamWriter, VideoWriter, FileVideoWriter
from platypush.plugins.camera.model.writer.preview import PreviewWriter


@dataclass
class CameraInfo:
    device: Optional[Union[int, str]]
    bind_address: Optional[str] = None
    capture_timeout: float = 20.0
    color_transform: Optional[str] = None
    ffmpeg_bin: Optional[str] = None
    fps: Optional[float] = None
    frames_dir: Optional[str] = None
    grayscale: Optional[bool] = None
    horizontal_flip: bool = False
    input_codec: Optional[str] = None
    input_format: Optional[str] = None
    listen_port: Optional[int] = None
    output_codec: Optional[str] = None
    output_format: Optional[str] = None
    resolution: Optional[Tuple[int, int]] = None
    rotate: Optional[float] = None
    scale_x: Optional[float] = None
    scale_y: Optional[float] = None
    stream_format: Optional[str] = None
    vertical_flip: bool = False
    warmup_frames: int = 0
    warmup_seconds: float = 0.

    def set(self, **kwargs):
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)

    def to_dict(self) -> dict:
        return {
            'bind_address': self.bind_address,
            'capture_timeout': self.capture_timeout,
            'color_transform': self.color_transform,
            'device': self.device,
            'ffmpeg_bin': self.ffmpeg_bin,
            'fps': self.fps,
            'frames_dir': self.frames_dir,
            'grayscale': self.grayscale,
            'horizontal_flip': self.horizontal_flip,
            'input_codec': self.input_codec,
            'input_format': self.input_format,
            'listen_port': self.listen_port,
            'output_codec': self.output_codec,
            'output_format': self.output_format,
            'resolution': list(self.resolution or ()),
            'rotate': self.rotate,
            'scale_x': self.scale_x,
            'scale_y': self.scale_y,
            'stream_format': self.stream_format,
            'vertical_flip': self.vertical_flip,
            'warmup_frames': self.warmup_frames,
            'warmup_seconds': self.warmup_seconds,
        }

    def clone(self):
        # noinspection PyArgumentList
        return self.__class__(**self.to_dict())


@dataclass
class Camera:
    info: CameraInfo
    start_event: threading.Event = threading.Event()
    stream_event: threading.Event = threading.Event()
    capture_thread: Optional[threading.Thread] = None
    stream_thread: Optional[threading.Thread] = None
    object = None
    stream: Optional[StreamWriter] = None
    preview: Optional[PreviewWriter] = None
    file_writer: Optional[FileVideoWriter] = None

    def get_outputs(self) -> Set[VideoWriter]:
        writers = set()
        # if self.preview and self.preview.is_alive():
        if self.preview and not self.preview.closed:
            writers.add(self.preview)

        if self.stream and not self.stream.closed:
            writers.add(self.stream)

        if self.file_writer and not self.file_writer.closed:
            writers.add(self.file_writer)

        return writers

    def effective_resolution(self) -> Tuple[int, int]:
        rot = (self.info.rotate or 0) * math.pi / 180
        sin = math.sin(rot)
        cos = math.cos(rot)
        scale = np.array([[self.info.scale_x or 1., self.info.scale_y or 1.]])
        resolution = np.array([[self.info.resolution[0], self.info.resolution[1]]])
        rot_matrix = np.array([[sin, cos], [cos, sin]])
        resolution = (scale * abs(np.cross(rot_matrix, resolution)))[0]
        return int(round(resolution[0])), int(round(resolution[1]))


# vim:sw=4:ts=4:et:



from platypush.context import get_plugin
from platypush.plugins.media.search import MediaSearcher


class JellyfinMediaSearcher(MediaSearcher):
    def search(self, query, **_):
        """
        Performs a search on a Jellyfin server using the configured
        :class:`platypush.plugins.media.jellyfin.MediaJellyfinPlugin`
        instance (if configured).
        """

        try:
            media = get_plugin('media.jellyfin')
        except RuntimeError:
            return []

        if not media:
            return []

        self.logger.info('Searching Jellyfin for "{}"'.format(query))
        results = media.search(query=query).output
        self.logger.info('{} Jellyfin results found for the search query "{}"'.format(len(results), query))
        return results


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.nextcloud import NextCloudActivityEvent
from platypush.plugins.nextcloud import NextcloudPlugin
from platypush.plugins.variable import VariablePlugin


class NextcloudBackend(Backend):
    """
    This backend triggers events when new activities occur on a NextCloud instance.

    Triggers:

        - :class:`platypush.message.event.nextcloud.NextCloudActivityEvent` when new activity occurs on the instance.
          The field ``activity_type`` identifies the activity type (e.g. ``file_created``, ``file_deleted``,
          ``file_changed``). Example in the case of the creation of new files:

          .. code-block:: json

            {
              "activity_id": 387,
              "app": "files",
              "activity_type": "file_created",
              "user": "your-user",
              "subject": "You created InstantUpload/Camera/IMG_0100.jpg, InstantUpload/Camera/IMG_0101.jpg and InstantUpload/Camera/IMG_0102.jpg",
              "subject_rich": [
                "You created {file3}, {file2} and {file1}",
                {
                  "file1": {
                    "type": "file",
                    "id": "41994",
                    "name": "IMG_0100.jpg",
                    "path": "InstantUpload/Camera/IMG_0100.jpg",
                    "link": "https://your-domain/nextcloud/index.php/f/41994"
                  },
                  "file2": {
                    "type": "file",
                    "id": "42005",
                    "name": "IMG_0101.jpg",
                    "path": "InstantUpload/Camera/IMG_0102.jpg",
                    "link": "https://your-domain/nextcloud/index.php/f/42005"
                  },
                  "file3": {
                    "type": "file",
                    "id": "42014",
                    "name": "IMG_0102.jpg",
                    "path": "InstantUpload/Camera/IMG_0102.jpg",
                    "link": "https://your-domain/nextcloud/index.php/f/42014"
                  }
                }
              ],
              "message": "",
              "message_rich": [
                "",
                []
              ],
              "object_type": "files",
              "object_id": 41994,
              "object_name": "/InstantUpload/Camera/IMG_0102.jpg",
              "objects": {
                "42014": "/InstantUpload/Camera/IMG_0100.jpg",
                "42005": "/InstantUpload/Camera/IMG_0101.jpg",
                "41994": "/InstantUpload/Camera/IMG_0102.jpg"
              },
              "link": "https://your-domain/nextcloud/index.php/apps/files/?dir=/InstantUpload/Camera",
              "icon": "https://your-domain/nextcloud/apps/files/img/add-color.svg",
              "datetime": "2020-09-07T17:04:29+00:00"
            }

    """

    _LAST_ACTIVITY_VARNAME = '_NEXTCLOUD_LAST_ACTIVITY_ID'

    def __init__(self, url: Optional[str] = None, username: Optional[str] = None, password: Optional[str] = None,
                 object_type: Optional[str] = None, object_id: Optional[int] = None,
                 poll_seconds: Optional[float] = 60., **kwargs):
        """
        :param url: NextCloud instance URL (default: same as the :class:`platypush.plugins.nextcloud.NextCloudPlugin`).
        :param username: NextCloud username (default: same as the :class:`platypush.plugins.nextcloud.NextCloudPlugin`).
        :param password: NextCloud password (default: same as the :class:`platypush.plugins.nextcloud.NextCloudPlugin`).
        :param object_type: If set, only filter events on this type of object.
        :param object_id: If set, only filter events on this object ID.
        :param poll_seconds: How often the backend should poll the instance (default: one minute).
        """
        super().__init__(**kwargs)
        self.url: Optional[str] = None
        self.username: Optional[str] = None
        self.password: Optional[str] = None
        self.object_type = object_type
        self.object_id = object_id
        self.poll_seconds = poll_seconds
        self._last_seen_id = None

        try:
            plugin: Optional[NextcloudPlugin] = get_plugin('nextcloud')
            if plugin:
                self.url = plugin.conf.url
                self.username = plugin.conf.username
                self.password = plugin.conf.password
        except Exception as e:
            self.logger.info('NextCloud plugin not configured: {}'.format(str(e)))

        self.url = url if url else self.url
        self.username = username if username else self.username
        self.password = password if password else self.password

        assert self.url and self.username and self.password, \
            'No configuration provided neither for the NextCloud plugin nor the backend'

    @property
    def last_seen_id(self) -> Optional[int]:
        if self._last_seen_id is None:
            variables: VariablePlugin = get_plugin('variable')
            last_seen_id = variables.get(self._LAST_ACTIVITY_VARNAME).output.get(self._LAST_ACTIVITY_VARNAME)
            self._last_seen_id = last_seen_id

        return self._last_seen_id

    @last_seen_id.setter
    def last_seen_id(self, value: Optional[int]):
        variables: VariablePlugin = get_plugin('variable')
        variables.set(**{self._LAST_ACTIVITY_VARNAME: value})
        self._last_seen_id = value

    @staticmethod
    def _activity_to_event(activity: dict) -> NextCloudActivityEvent:
        return NextCloudActivityEvent(activity_type=activity.pop('type'), **activity)

    def loop(self):
        last_seen_id = int(self.last_seen_id)
        new_last_seen_id = int(last_seen_id)
        plugin: NextcloudPlugin = get_plugin('nextcloud')
        # noinspection PyUnresolvedReferences
        activities = plugin.get_activities(sort='desc', url=self.url, username=self.username, password=self.password,
                                           object_type=self.object_type, object_id=self.object_id).output

        events = []
        for activity in activities:
            if last_seen_id and activity['activity_id'] <= last_seen_id:
                break

            events.append(self._activity_to_event(activity))

            if not new_last_seen_id or activity['activity_id'] > new_last_seen_id:
                new_last_seen_id = int(activity['activity_id'])

        for evt in events[::-1]:
            self.bus.post(evt)

        if new_last_seen_id and last_seen_id != new_last_seen_id:
            self.last_seen_id = new_last_seen_id


# vim:sw=4:ts=4:et:



import asyncio
import datetime
import json
import logging
import os
import pathlib
import threading

from dataclasses import dataclass
from typing import Collection, Dict, Optional, Union
from urllib.parse import urlparse

from async_lru import alru_cache
from nio import (
    AsyncClient,
    AsyncClientConfig,
    CallAnswerEvent,
    CallHangupEvent,
    CallInviteEvent,
    Event,
    InviteEvent,
    KeyVerificationStart,
    KeyVerificationAccept,
    KeyVerificationMac,
    KeyVerificationKey,
    KeyVerificationCancel,
    LocalProtocolError,
    LoginResponse,
    MatrixRoom,
    MegolmEvent,
    ProfileGetResponse,
    RoomCreateEvent,
    RoomEncryptedAudio,
    RoomEncryptedFile,
    RoomEncryptedImage,
    RoomEncryptedMedia,
    RoomEncryptedVideo,
    RoomGetEventError,
    RoomGetStateResponse,
    RoomMemberEvent,
    RoomMessageAudio,
    RoomMessageFile,
    RoomMessageFormatted,
    RoomMessageText,
    RoomMessageImage,
    RoomMessageMedia,
    RoomMessageVideo,
    RoomTopicEvent,
    RoomUpgradeEvent,
    StickerEvent,
    SyncResponse,
    ToDeviceError,
    UnknownEncryptedEvent,
    UnknownEvent,
)

import aiofiles
import aiofiles.os

from nio.client.async_client import client_session
from nio.client.base_client import logged_in
from nio.crypto import decrypt_attachment
from nio.crypto.device import OlmDevice
from nio.events.ephemeral import ReceiptEvent, TypingNoticeEvent
from nio.events.presence import PresenceEvent
from nio.responses import DownloadResponse, RoomMessagesResponse

from platypush.config import Config
from platypush.context import get_bus
from platypush.message.event.matrix import (
    MatrixCallAnswerEvent,
    MatrixCallHangupEvent,
    MatrixCallInviteEvent,
    MatrixEncryptedMessageEvent,
    MatrixMessageAudioEvent,
    MatrixMessageEvent,
    MatrixMessageFileEvent,
    MatrixMessageImageEvent,
    MatrixMessageVideoEvent,
    MatrixReactionEvent,
    MatrixRoomCreatedEvent,
    MatrixRoomInviteEvent,
    MatrixRoomJoinEvent,
    MatrixRoomLeaveEvent,
    MatrixRoomSeenReceiptEvent,
    MatrixRoomTopicChangedEvent,
    MatrixRoomTypingStartEvent,
    MatrixRoomTypingStopEvent,
    MatrixSyncEvent,
    MatrixUserPresenceEvent,
)

from platypush.utils import get_mime_type

logger = logging.getLogger(__name__)


@dataclass
class Credentials:
    server_url: str
    user_id: str
    access_token: str
    device_id: str | None

    def to_dict(self) -> dict:
        return {
            'server_url': self.server_url,
            'user_id': self.user_id,
            'access_token': self.access_token,
            'device_id': self.device_id,
        }


class MatrixClient(AsyncClient):
    def __init__(
        self,
        *args,
        credentials_file: str,
        store_path: str | None = None,
        config: Optional[AsyncClientConfig] = None,
        autojoin_on_invite=True,
        autotrust_devices=False,
        autotrust_devices_whitelist: Collection[str] | None = None,
        autotrust_rooms_whitelist: Collection[str] | None = None,
        autotrust_users_whitelist: Collection[str] | None = None,
        **kwargs,
    ):
        credentials_file = os.path.abspath(os.path.expanduser(credentials_file))

        if not store_path:
            store_path = os.path.join(Config.get('workdir'), 'matrix', 'store')  # type: ignore

        assert store_path
        store_path = os.path.abspath(os.path.expanduser(store_path))
        pathlib.Path(store_path).mkdir(exist_ok=True, parents=True)

        if not config:
            config = AsyncClientConfig(
                max_limit_exceeded=0,
                max_timeouts=0,
                store_sync_tokens=True,
                encryption_enabled=True,
            )

        super().__init__(*args, config=config, store_path=store_path, **kwargs)
        self.logger = logging.getLogger(self.__class__.__name__)
        self._credentials_file = credentials_file
        self._autojoin_on_invite = autojoin_on_invite
        self._autotrust_devices = autotrust_devices
        self._autotrust_devices_whitelist = autotrust_devices_whitelist
        self._autotrust_rooms_whitelist = autotrust_rooms_whitelist or set()
        self._autotrust_users_whitelist = autotrust_users_whitelist or set()
        self._first_sync_performed = asyncio.Event()
        self._last_batches_by_room = {}
        self._typing_users_by_room = {}

        self._encrypted_attachments_keystore_path = os.path.join(
            store_path, 'attachment_keys.json'
        )
        self._encrypted_attachments_keystore = {}
        self._sync_store_timer: threading.Timer | None = None
        keystore = {}

        try:
            with open(self._encrypted_attachments_keystore_path, 'r') as f:
                keystore = json.load(f)
        except (ValueError, OSError):
            with open(self._encrypted_attachments_keystore_path, 'w') as f:
                f.write(json.dumps({}))

        pathlib.Path(self._encrypted_attachments_keystore_path).touch(
            mode=0o600, exist_ok=True
        )

        self._encrypted_attachments_keystore = {
            tuple(key.split('|')): data for key, data in keystore.items()
        }

    async def _autojoin_room_callback(self, room: MatrixRoom, *_):
        await self.join(room.room_id)  # type: ignore

    def _load_from_file(self):
        if not os.path.isfile(self._credentials_file):
            return

        try:
            with open(self._credentials_file, 'r') as f:
                credentials = json.load(f)
        except json.JSONDecodeError:
            self.logger.warning(
                'Could not read credentials_file %s - overwriting it',
                self._credentials_file,
            )
            return

        assert credentials.get('user_id'), 'Missing user_id'
        assert credentials.get('access_token'), 'Missing access_token'

        self.access_token = credentials['access_token']
        self.user_id = credentials['user_id']
        self.homeserver = credentials.get('server_url', self.homeserver)
        if credentials.get('device_id'):
            self.device_id = credentials['device_id']

        self.load_store()

    async def login(
        self,
        password: str | None = None,
        device_name: str | None = None,
        token: str | None = None,
    ) -> LoginResponse:
        self._load_from_file()
        login_res = None

        if self.access_token:
            self.load_store()
            self.logger.info(
                'Logged in to %s as %s using the stored access token',
                self.homeserver,
                self.user_id,
            )

            login_res = LoginResponse(
                user_id=self.user_id,
                device_id=self.device_id,
                access_token=self.access_token,
            )
        else:
            assert self.user, 'No credentials file found and no user provided'
            login_args = {'device_name': device_name}
            if token:
                login_args['token'] = token
            else:
                assert (
                    password
                ), 'No credentials file found and no password nor access token provided'
                login_args['password'] = password

            login_res = await super().login(**login_args)
            assert isinstance(login_res, LoginResponse), f'Failed to login: {login_res}'
            self.logger.info(login_res)

            credentials = Credentials(
                server_url=self.homeserver,
                user_id=login_res.user_id,
                access_token=login_res.access_token,
                device_id=login_res.device_id,
            )

            with open(self._credentials_file, 'w') as f:
                json.dump(credentials.to_dict(), f)
            os.chmod(self._credentials_file, 0o600)

        if self.should_upload_keys:
            self.logger.info('Uploading encryption keys')
            await self.keys_upload()

        self.logger.info('Synchronizing state')
        self._first_sync_performed.clear()
        self._add_callbacks()
        sync_token = self.loaded_sync_token
        self.loaded_sync_token = ''
        await self.sync(sync_filter={'room': {'timeline': {'limit': 1}}})
        self.loaded_sync_token = sync_token

        self._sync_devices_trust()
        self._first_sync_performed.set()

        get_bus().post(MatrixSyncEvent(server_url=self.homeserver))
        self.logger.info('State synchronized')
        return login_res

    @logged_in
    async def sync(self, *args, **kwargs) -> SyncResponse:
        response = await super().sync(*args, **kwargs)
        assert isinstance(response, SyncResponse), str(response)
        self._last_batches_by_room.update(
            {
                room_id: {
                    'prev_batch': room.timeline.prev_batch,
                    'next_batch': response.next_batch,
                }
                for room_id, room in response.rooms.join.items()
            }
        )

        return response

    @logged_in
    async def room_messages(
        self, room_id: str, start: str | None = None, *args, **kwargs
    ) -> RoomMessagesResponse:
        if not start:
            start = self._last_batches_by_room.get(room_id, {}).get('prev_batch')
            assert start, (
                f'No sync batches were found for room {room_id} and no start'
                'batch has been provided'
            )

        response = await super().room_messages(room_id, start, *args, **kwargs)
        assert isinstance(response, RoomMessagesResponse), str(response)
        return response

    def _sync_devices_trust(self):
        all_devices = self.get_devices()
        devices_to_trust: Dict[str, OlmDevice] = {}
        untrusted_devices = {
            device_id: device
            for device_id, device in all_devices.items()
            if not device.verified
        }

        if self._autotrust_devices:
            devices_to_trust.update(untrusted_devices)
        else:
            if self._autotrust_devices_whitelist:
                devices_to_trust.update(
                    {
                        device_id: device
                        for device_id, device in all_devices.items()
                        if device_id in self._autotrust_devices_whitelist
                        and device_id in untrusted_devices
                    }
                )
            if self._autotrust_rooms_whitelist:
                devices_to_trust.update(
                    {
                        device_id: device
                        for room_id, devices in self.get_devices_by_room().items()
                        for device_id, device in devices.items()  # type: ignore
                        if room_id in self._autotrust_rooms_whitelist
                        and device_id in untrusted_devices
                    }
                )
            if self._autotrust_users_whitelist:
                devices_to_trust.update(
                    {
                        device_id: device
                        for user_id, devices in self.get_devices_by_user().items()
                        for device_id, device in devices.items()  # type: ignore
                        if user_id in self._autotrust_users_whitelist
                        and device_id in untrusted_devices
                    }
                )

        for device in devices_to_trust.values():
            self.verify_device(device)
            self.logger.info(
                'Device %s by user %s added to the whitelist', device.id, device.user_id
            )

    def get_devices_by_user(
        self, user_id: str | None = None
    ) -> Dict[str, Dict[str, OlmDevice]] | Dict[str, OlmDevice]:
        devices = {user: devices for user, devices in self.device_store.items()}

        if user_id:
            devices = devices.get(user_id, {})
        return devices

    def get_devices(self) -> Dict[str, OlmDevice]:
        return {
            device_id: device
            for _, devices in self.device_store.items()
            for device_id, device in devices.items()
        }

    def get_device(self, device_id: str) -> Optional[OlmDevice]:
        return self.get_devices().get(device_id)

    def get_devices_by_room(
        self, room_id: str | None = None
    ) -> Dict[str, Dict[str, OlmDevice]] | Dict[str, OlmDevice]:
        devices = {
            room_id: {
                device_id: device
                for _, devices in self.room_devices(room_id).items()
                for device_id, device in devices.items()
            }
            for room_id in self.rooms.keys()
        }

        if room_id:
            devices = devices.get(room_id, {})
        return devices

    def _add_callbacks(self):
        self.add_event_callback(self._event_catch_all, Event)
        self.add_event_callback(self._on_invite, InviteEvent)  # type: ignore
        self.add_event_callback(self._on_message, RoomMessageText)  # type: ignore
        self.add_event_callback(self._on_message, RoomMessageMedia)  # type: ignore
        self.add_event_callback(self._on_message, RoomEncryptedMedia)  # type: ignore
        self.add_event_callback(self._on_message, StickerEvent)  # type: ignore
        self.add_event_callback(self._on_room_member, RoomMemberEvent)  # type: ignore
        self.add_event_callback(self._on_room_topic_changed, RoomTopicEvent)  # type: ignore
        self.add_event_callback(self._on_call_invite, CallInviteEvent)  # type: ignore
        self.add_event_callback(self._on_call_answer, CallAnswerEvent)  # type: ignore
        self.add_event_callback(self._on_call_hangup, CallHangupEvent)  # type: ignore
        self.add_event_callback(self._on_unknown_event, UnknownEvent)  # type: ignore
        self.add_event_callback(self._on_unknown_encrypted_event, UnknownEncryptedEvent)  # type: ignore
        self.add_event_callback(self._on_unknown_encrypted_event, MegolmEvent)  # type: ignore
        self.add_to_device_callback(self._on_key_verification_start, KeyVerificationStart)  # type: ignore
        self.add_to_device_callback(self._on_key_verification_cancel, KeyVerificationCancel)  # type: ignore
        self.add_to_device_callback(self._on_key_verification_key, KeyVerificationKey)  # type: ignore
        self.add_to_device_callback(self._on_key_verification_mac, KeyVerificationMac)  # type: ignore
        self.add_to_device_callback(self._on_key_verification_accept, KeyVerificationAccept)  # type: ignore
        self.add_ephemeral_callback(self._on_typing, TypingNoticeEvent)  # type: ignore
        self.add_ephemeral_callback(self._on_receipt, ReceiptEvent)  # type: ignore
        self.add_presence_callback(self._on_presence, PresenceEvent)  # type: ignore

        if self._autojoin_on_invite:
            self.add_event_callback(self._autojoin_room_callback, InviteEvent)  # type: ignore

    def _sync_store(self):
        self.logger.info('Synchronizing keystore')
        serialized_keystore = json.dumps(
            {
                f'{server}|{media_id}': data
                for (
                    server,
                    media_id,
                ), data in self._encrypted_attachments_keystore.items()
            }
        )

        try:
            with open(self._encrypted_attachments_keystore_path, 'w') as f:
                f.write(serialized_keystore)
        finally:
            self._sync_store_timer = None

    @alru_cache(maxsize=500)
    @client_session
    async def get_profile(self, user_id: str | None = None) -> ProfileGetResponse:
        """
        Cached version of get_profile.
        """
        ret = await super().get_profile(user_id)
        assert isinstance(
            ret, ProfileGetResponse
        ), f'Could not retrieve profile for user {user_id}: {ret.message}'
        return ret

    @alru_cache(maxsize=500)
    @client_session
    async def room_get_state(self, room_id: str) -> RoomGetStateResponse:
        """
        Cached version of room_get_state.
        """
        ret = await super().room_get_state(room_id)
        assert isinstance(
            ret, RoomGetStateResponse
        ), f'Could not retrieve profile for room {room_id}: {ret.message}'
        return ret

    @client_session
    async def download(
        self,
        server_name: str,
        media_id: str,
        filename: str | None = None,
        allow_remote: bool = True,
    ):
        response = await super().download(
            server_name, media_id, filename, allow_remote=allow_remote
        )

        assert isinstance(
            response, DownloadResponse
        ), f'Could not download media {media_id}: {response}'

        encryption_data = self._encrypted_attachments_keystore.get(
            (server_name, media_id)
        )
        if encryption_data:
            self.logger.info('Decrypting media %s using the available keys', media_id)
            response.filename = encryption_data.get('body', response.filename)
            response.content_type = encryption_data.get(
                'mimetype', response.content_type
            )
            response.body = decrypt_attachment(
                response.body,
                key=encryption_data.get('key'),
                hash=encryption_data.get('hash'),
                iv=encryption_data.get('iv'),
            )

        return response

    async def _event_base_args(
        self, room: Optional[MatrixRoom], event: Optional[Event] = None
    ) -> dict:
        sender_id = getattr(event, 'sender', None)
        sender = (
            await self.get_profile(sender_id) if sender_id else None  # type: ignore
        )

        return {
            'server_url': self.homeserver,
            'sender_id': sender_id,
            'sender_display_name': sender.displayname if sender else None,
            'sender_avatar_url': sender.avatar_url if sender else None,
            **(
                {
                    'room_id': room.room_id,
                    'room_name': room.name,
                    'room_topic': room.topic,
                }
                if room
                else {}
            ),
            'server_timestamp': (
                datetime.datetime.fromtimestamp(event.server_timestamp / 1000)
                if event and getattr(event, 'server_timestamp', None)
                else None
            ),
        }

    async def _event_catch_all(self, room: MatrixRoom, event: Event):
        self.logger.debug('Received event on room %s: %r', room.room_id, event)

    async def _on_invite(self, room: MatrixRoom, event: RoomMessageText):
        get_bus().post(
            MatrixRoomInviteEvent(
                **(await self._event_base_args(room, event)),
            )
        )

    async def _on_message(
        self,
        room: MatrixRoom,
        event: Union[
            RoomMessageText, RoomMessageMedia, RoomEncryptedMedia, StickerEvent
        ],
    ):
        if self._first_sync_performed.is_set():
            evt_type = MatrixMessageEvent
            evt_args = {
                'body': event.body,
                'url': getattr(event, 'url', None),
                **(await self._event_base_args(room, event)),
            }

            if isinstance(event, (RoomMessageMedia, RoomEncryptedMedia, StickerEvent)):
                evt_args['url'] = event.url

            if isinstance(event, RoomEncryptedMedia):
                evt_args['thumbnail_url'] = event.thumbnail_url
                evt_args['mimetype'] = event.mimetype
                self._store_encrypted_media_keys(event)
            if isinstance(event, RoomMessageFormatted):
                evt_args['format'] = event.format
                evt_args['formatted_body'] = event.formatted_body

            if isinstance(event, (RoomMessageImage, RoomEncryptedImage)):
                evt_type = MatrixMessageImageEvent
            elif isinstance(event, (RoomMessageAudio, RoomEncryptedAudio)):
                evt_type = MatrixMessageAudioEvent
            elif isinstance(event, (RoomMessageVideo, RoomEncryptedVideo)):
                evt_type = MatrixMessageVideoEvent
            elif isinstance(event, (RoomMessageFile, RoomEncryptedFile)):
                evt_type = MatrixMessageFileEvent

            get_bus().post(evt_type(**evt_args))

    def _store_encrypted_media_keys(self, event: RoomEncryptedMedia):
        url = event.url.strip('/')
        parsed_url = urlparse(url)
        homeserver = parsed_url.netloc.strip('/')
        media_key = (homeserver, parsed_url.path.strip('/'))

        self._encrypted_attachments_keystore[media_key] = {
            'url': url,
            'body': event.body,
            'key': event.key['k'],
            'hash': event.hashes['sha256'],
            'iv': event.iv,
            'homeserver': homeserver,
            'mimetype': event.mimetype,
        }

        if not self._sync_store_timer:
            self._sync_store_timer = threading.Timer(5, self._sync_store)
            self._sync_store_timer.start()

    async def _on_room_member(self, room: MatrixRoom, event: RoomMemberEvent):
        evt_type = None
        if event.membership == 'join':
            evt_type = MatrixRoomJoinEvent
        elif event.membership == 'leave':
            evt_type = MatrixRoomLeaveEvent

        if evt_type and self._first_sync_performed.is_set():
            get_bus().post(
                evt_type(
                    **(await self._event_base_args(room, event)),
                )
            )

    async def _on_room_topic_changed(self, room: MatrixRoom, event: RoomTopicEvent):
        if self._first_sync_performed.is_set():
            get_bus().post(
                MatrixRoomTopicChangedEvent(
                    **(await self._event_base_args(room, event)),
                    topic=event.topic,
                )
            )

    async def _on_call_invite(self, room: MatrixRoom, event: CallInviteEvent):
        if self._first_sync_performed.is_set():
            get_bus().post(
                MatrixCallInviteEvent(
                    call_id=event.call_id,
                    version=event.version,
                    invite_validity=event.lifetime / 1000.0,
                    sdp=event.offer.get('sdp'),
                    **(await self._event_base_args(room, event)),
                )
            )

    async def _on_call_answer(self, room: MatrixRoom, event: CallAnswerEvent):
        if self._first_sync_performed.is_set():
            get_bus().post(
                MatrixCallAnswerEvent(
                    call_id=event.call_id,
                    version=event.version,
                    sdp=event.answer.get('sdp'),
                    **(await self._event_base_args(room, event)),
                )
            )

    async def _on_call_hangup(self, room: MatrixRoom, event: CallHangupEvent):
        if self._first_sync_performed.is_set():
            get_bus().post(
                MatrixCallHangupEvent(
                    call_id=event.call_id,
                    version=event.version,
                    **(await self._event_base_args(room, event)),
                )
            )

    async def _on_room_created(self, room: MatrixRoom, event: RoomCreateEvent):
        get_bus().post(
            MatrixRoomCreatedEvent(
                **(await self._event_base_args(room, event)),
            )
        )

    def _get_sas(self, event):
        sas = self.key_verifications.get(event.transaction_id)
        if not sas:
            self.logger.debug(
                'Received a key verification event with no associated transaction ID'
            )

        return sas

    async def _on_key_verification_start(self, event: KeyVerificationStart):
        self.logger.info(f'Received a key verification request from {event.sender}')

        if 'emoji' not in event.short_authentication_string:
            self.logger.warning(
                'Only emoji verification is supported, but the verifying device '
                'provided the following authentication methods: %r',
                event.short_authentication_string,
            )
            return

        sas = self._get_sas(event)
        if not sas:
            return

        rs = await self.accept_key_verification(sas.transaction_id)
        assert not isinstance(
            rs, ToDeviceError
        ), f'accept_key_verification failed: {rs}'

        rs = await self.to_device(sas.share_key())
        assert not isinstance(rs, ToDeviceError), f'Shared key exchange failed: {rs}'

    async def _on_key_verification_accept(self, event: KeyVerificationAccept):
        self.logger.info('Key verification from device %s accepted', event.sender)

    async def _on_key_verification_cancel(self, event: KeyVerificationCancel):
        self.logger.info(
            'The device %s cancelled a key verification request. ' 'Reason: %s',
            event.sender,
            event.reason,
        )

    async def _on_key_verification_key(self, event: KeyVerificationKey):
        sas = self._get_sas(event)
        if not sas:
            return

        self.logger.info(
            'Received emoji verification from device %s: %s',
            event.sender,
            sas.get_emoji(),
        )

        rs = await self.confirm_short_auth_string(sas.transaction_id)
        assert not isinstance(
            rs, ToDeviceError
        ), f'confirm_short_auth_string failed: {rs}'

    async def _on_key_verification_mac(self, event: KeyVerificationMac):
        self.logger.info('Received MAC verification request from %s', event.sender)
        sas = self._get_sas(event)
        if not sas:
            return

        try:
            mac = sas.get_mac()
        except LocalProtocolError as e:
            self.logger.warning(
                'Verification from %s cancelled or unexpected protocol error. '
                'Reason: %s',
                e,
                event.sender,
            )
            return

        rs = await self.to_device(mac)
        assert not isinstance(
            rs, ToDeviceError
        ), f'Sending of the verification MAC to {event.sender} failed: {rs}'

        self.logger.info('This device has been successfully verified!')

    async def _on_room_upgrade(self, room: MatrixRoom, event: RoomUpgradeEvent):
        self.logger.info(
            'The room %s has been moved to %s', room.room_id, event.replacement_room
        )

        await self.room_leave(room.room_id)
        await self.join(event.replacement_room)

    async def _on_typing(self, room: MatrixRoom, event: TypingNoticeEvent):
        users = set(event.users)
        typing_users = self._typing_users_by_room.get(room.room_id, set())
        start_typing_users = users.difference(typing_users)
        stop_typing_users = typing_users.difference(users)

        for user in start_typing_users:
            event.sender = user  # type: ignore
            get_bus().post(
                MatrixRoomTypingStartEvent(
                    **(await self._event_base_args(room, event)),  # type: ignore
                    sender=user,
                )
            )

        for user in stop_typing_users:
            event.sender = user  # type: ignore
            get_bus().post(
                MatrixRoomTypingStopEvent(
                    **(await self._event_base_args(room, event)),  # type: ignore
                )
            )

        self._typing_users_by_room[room.room_id] = users

    async def _on_receipt(self, room: MatrixRoom, event: ReceiptEvent):
        if self._first_sync_performed.is_set():
            for receipt in event.receipts:
                event.sender = receipt.user_id  # type: ignore
                get_bus().post(
                    MatrixRoomSeenReceiptEvent(
                        **(await self._event_base_args(room, event)),  # type: ignore
                    )
                )

    async def _on_presence(self, event: PresenceEvent):
        if self._first_sync_performed.is_set():
            last_active = (
                (
                    datetime.datetime.now()
                    - datetime.timedelta(seconds=event.last_active_ago / 1000)
                )
                if event.last_active_ago
                else None
            )

            event.sender = event.user_id  # type: ignore
            get_bus().post(
                MatrixUserPresenceEvent(
                    **(await self._event_base_args(None, event)),  # type: ignore
                    is_active=event.currently_active or False,
                    last_active=last_active,
                )
            )

    async def _on_unknown_encrypted_event(
        self, room: MatrixRoom, event: Union[UnknownEncryptedEvent, MegolmEvent]
    ):
        if self._first_sync_performed.is_set():
            body = getattr(event, 'ciphertext', '')
            get_bus().post(
                MatrixEncryptedMessageEvent(
                    body=body,
                    **(await self._event_base_args(room, event)),
                )
            )

    async def _on_unknown_event(self, room: MatrixRoom, event: UnknownEvent):
        evt = None

        if event.type == 'm.reaction' and self._first_sync_performed.is_set():
            # Get the ID of the event this was a reaction to
            relation_dict = event.source.get('content', {}).get('m.relates_to', {})
            reacted_to = relation_dict.get('event_id')
            if reacted_to and relation_dict.get('rel_type') == 'm.annotation':
                event_response = await self.room_get_event(room.room_id, reacted_to)

                if isinstance(event_response, RoomGetEventError):
                    self.logger.warning(
                        'Error getting event that was reacted to (%s)', reacted_to
                    )
                else:
                    evt = MatrixReactionEvent(
                        in_response_to_event_id=event_response.event.event_id,
                        **(await self._event_base_args(room, event)),
                    )

        if evt:
            get_bus().post(evt)
        else:
            self.logger.info(
                'Received an unknown event on room %s: %r', room.room_id, event
            )

    async def upload_file(
        self,
        file: str,
        name: Optional[str] = None,
        content_type: Optional[str] = None,
        encrypt: bool = False,
    ):
        file = os.path.expanduser(file)
        file_stat = await aiofiles.os.stat(file)

        async with aiofiles.open(file, 'rb') as f:
            return await super().upload(
                f,  # type: ignore
                content_type=(
                    content_type or get_mime_type(file) or 'application/octet-stream'
                ),
                filename=name or os.path.basename(file),
                encrypt=encrypt,
                filesize=file_stat.st_size,
            )


# vim:sw=4:ts=4:et:



import os

from platypush.backend import Backend
from platypush.message.event.inotify import InotifyCreateEvent, InotifyDeleteEvent, \
    InotifyOpenEvent, InotifyModifyEvent, InotifyCloseEvent, InotifyAccessEvent, InotifyMovedEvent


class InotifyBackend(Backend):
    """
    **NOTE**: This backend is *deprecated* in favour of :class:`platypush.backend.file.monitor.FileMonitorBackend`.

    (Linux only) This backend will listen for events on the filesystem (whether
    a file/directory on a watch list is opened, modified, created, deleted,
    closed or had its permissions changed) and will trigger a relevant event.

    Triggers:

        * :class:`platypush.message.event.inotify.InotifyCreateEvent` if a resource is created
        * :class:`platypush.message.event.inotify.InotifyAccessEvent` if a resource is accessed
        * :class:`platypush.message.event.inotify.InotifyOpenEvent` if a resource is opened
        * :class:`platypush.message.event.inotify.InotifyModifyEvent` if a resource is modified
        * :class:`platypush.message.event.inotify.InotifyPermissionsChangeEvent` if the permissions of a resource are changed
        * :class:`platypush.message.event.inotify.InotifyCloseEvent` if a resource is closed
        * :class:`platypush.message.event.inotify.InotifyDeleteEvent` if a resource is removed

    Requires:

        * **inotify** (``pip install inotify``)

    """

    inotify_watch = None

    def __init__(self, watch_paths=None, **kwargs):
        """
        :param watch_paths: Filesystem resources to watch for events
        :type watch_paths: str
        """

        super().__init__(**kwargs)
        self.watch_paths = set(map(
            lambda path: os.path.abspath(os.path.expanduser(path)),
            watch_paths if watch_paths else []))

    def _cleanup(self):
        if not self.inotify_watch:
            return

        for path in self.watch_paths:
            self.inotify_watch.remove_watch(path)

        self.inotify_watch = None

    def run(self):
        import inotify.adapters
        super().run()

        self.inotify_watch = inotify.adapters.Inotify()
        for path in self.watch_paths:
            self.inotify_watch.add_watch(path)

        moved_file = None
        self.logger.info('Initialized inotify file monitoring backend, monitored resources: {}'
                         .format(self.watch_paths))

        try:
            for inotify_event in self.inotify_watch.event_gen():
                if inotify_event is not None:
                    (header, inotify_types, watch_path, filename) = inotify_event
                    event = None
                    resource_type = inotify_types[1] if len(inotify_types) > 1 else None

                    if moved_file:
                        new = filename if 'IN_MOVED_TO' in inotify_types else None
                        event = InotifyMovedEvent(path=watch_path, old=moved_file, new=new)
                        moved_file = None

                    if 'IN_OPEN' in inotify_types:
                        event = InotifyOpenEvent(path=watch_path, resource=filename, resource_type=resource_type)
                    elif 'IN_ACCESS' in inotify_types:
                        event = InotifyAccessEvent(path=watch_path, resource=filename, resource_type=resource_type)
                    elif 'IN_CREATE' in inotify_types:
                        event = InotifyCreateEvent(path=watch_path, resource=filename, resource_type=resource_type)
                    elif 'IN_MOVED_FROM' in inotify_types:
                        moved_file = filename
                    elif 'IN_MOVED_TO' in inotify_types and not moved_file:
                        event = InotifyMovedEvent(path=watch_path, old=None, new=filename)
                    elif 'IN_DELETE' in inotify_types:
                        event = InotifyDeleteEvent(path=watch_path, resource=filename, resource_type=resource_type)
                    elif 'IN_MODIFY' in inotify_types:
                        event = InotifyModifyEvent(path=watch_path, resource=filename, resource_type=resource_type)
                    elif 'IN_CLOSE_WRITE' in inotify_types or 'IN_CLOSE_NOWRITE' in inotify_types:
                        event = InotifyCloseEvent(path=watch_path, resource=filename, resource_type=resource_type)

                    if event:
                        self.bus.post(event)
        finally:
            self._cleanup()


# vim:sw=4:ts=4:et:



import os
import requests
from typing import Optional

from platypush.message import Message
from platypush.message.response import Response
from platypush.utils import set_timeout, clear_timeout

from . import test_dir

# Default configuration folder for tests
conf_dir = os.path.join(test_dir, 'etc')

# Default configuration file for tests
config_file = os.path.join(conf_dir, 'config_test.yaml')

# Default request timeout in seconds
request_timeout = 10

# Default test user
test_user = 'platypush'

# Default test password
test_pass = 'test'

# Base URL
base_url = None


def set_base_url(url: str):
    global base_url
    base_url = url


class TimeoutException(RuntimeError):
    """
    Exception raised in case of timeout.
    """
    def __init__(self, msg: str = 'Timeout'):
        self.msg = msg


def send_request(action: str, timeout: Optional[float] = None, args: Optional[dict] = None,
                 parse_json: bool = True, authenticate: bool = True, **kwargs):
    if not timeout:
        timeout = request_timeout
    if not args:
        args = {}

    auth = (test_user, test_pass) if authenticate else kwargs.pop('auth', ())
    set_timeout(seconds=timeout, on_timeout=on_timeout('Receiver response timed out'))
    response = requests.post(
        '{}/execute'.format(base_url),
        auth=auth,
        json={
            'type': 'request',
            'action': action,
            'args': args,
        }, **kwargs
    )

    clear_timeout()

    if parse_json:
        response = parse_response(response)
    return response


def register_user(username: Optional[str] = None, password: Optional[str] = None):
    if not username:
        username = test_user
        password = test_pass

    set_timeout(seconds=request_timeout, on_timeout=on_timeout('User registration response timed out'))
    response = requests.post('{base_url}/register?redirect={base_url}/'.format(base_url=base_url), data={
        'username': username,
        'password': password,
        'confirm_password': password,
    })

    clear_timeout()
    return response


def on_timeout(msg):
    def _f(): raise TimeoutException(msg)
    return _f


def parse_response(response):
    response = Message.build(response.json())
    assert isinstance(response, Response), 'Expected Response type, got {}'.format(response.__class__.__name__)
    return response



import urllib.parse
from typing import Optional

from platypush.config import Config
from platypush.context import get_plugin
from platypush.plugins import Plugin, action
from platypush.plugins.media import MediaPlugin


class TtsPlugin(Plugin):
    """
    Default Text-to-Speech plugin. It leverages Google Translate.

    Requires:

        * At least a *media plugin* (see :class:`platypush.plugins.media.MediaPlugin`) enabled/configured - used for
          speech playback.

    """

    _supported_media_plugins = [
        'media.gstreamer',
        'media.omxplayer',
        'media.mplayer',
        'media.mpv',
        'media.vlc',
    ]

    def __init__(self, language='en-gb', media_plugin: Optional[str] = None, player_args: Optional[dict] = None):
        """
        :param language: Language code (default: ``en-gb``).
        :param media_plugin: Media plugin to be used for audio playback. Supported:

            - ``media.gstreamer``
            - ``media.omxplayer``
            - ``media.mplayer``
            - ``media.mpv``
            - ``media.vlc``

        :param player_args: Optional arguments that should be passed to the player plugin's
            :meth:`platypush.plugins.media.MediaPlugin.play` method.
        """
        super().__init__()
        self.language = language
        self.player_args = player_args or {}
        self.media_plugin = get_plugin(media_plugin) if media_plugin else self._get_media_plugin()
        assert self.media_plugin, 'No media playback plugin configured. Supported plugins: [{}]'.format(
            ', '.join(self._supported_media_plugins))

    @classmethod
    def _get_media_plugin(cls) -> Optional[MediaPlugin]:
        for plugin in cls._supported_media_plugins:
            if plugin in Config.get():
                return get_plugin(plugin)

    @action
    def say(self, text: str, language: Optional[str] = None, player_args: Optional[dict] = None):
        """
        Say some text.

        :param text: Text to say.
        :param language: Language code override.
        :param player_args: Optional arguments that should be passed to the player plugin's
            :meth:`platypush.plugins.media.MediaPlugin.play` method.
        """
        language = language or self.language
        player_args = player_args or self.player_args
        url = 'https://translate.google.com/translate_tts?{}'.format(
            urllib.parse.urlencode({
                'ie': 'UTF-8',
                'client': 'tw-ob',
                'tl': language,
                'q': text,
            }))

        self.media_plugin.play(url, **player_args)


# vim:sw=4:ts=4:et:



import json
import os
import re
import subprocess
import time
from typing import Optional, List

from platypush.backend import Backend
from platypush.message.event.joystick import JoystickConnectedEvent, JoystickDisconnectedEvent, JoystickStateEvent, \
    JoystickButtonPressedEvent, JoystickButtonReleasedEvent, JoystickAxisEvent


class JoystickState:
    def __init__(self, axes: List[int], buttons: List[bool]):
        self.axes = axes
        self.buttons = buttons

    def __str__(self):
        return json.dumps(self.__dict__)

    def __eq__(self, obj):
        return obj.axes == self.axes and obj.buttons == self.buttons

    def __sub__(self, obj) -> dict:
        if len(obj.axes) < len(self.axes) or len(obj.buttons) < len(self.buttons):
            return {}

        diff = {
            'axes': {
                axis: obj.axes[axis]
                for axis in range(len(self.axes))
                if self.axes[axis] != obj.axes[axis]
            },
            'buttons': {
                button: obj.buttons[button]
                for button in range(len(self.buttons))
                if self.buttons[button] != obj.buttons[button]
            },
        }

        return {
            k: v for k, v in diff.items() if v
        }


class JoystickJstestBackend(Backend):
    """
    This backend can be used to intercept events from a joystick device if the device does not work with the standard
    :class:`platypush.backend.joystick.JoystickBackend` backend (this may especially happen with some Bluetooth
    joysticks that don't support the ``ioctl`` requests used by ``inputs``).

    This backend only works on Linux and it requires the ``joystick`` package to be installed.

    **NOTE**: This backend can be quite slow, since it has to run another program (``jstest``) and parse its output.
    Consider it as a last resort if your joystick works with neither :class:`platypush.backend.joystick.JoystickBackend`
    nor :class:`platypush.backend.joystick.JoystickLinuxBackend`.

    Instructions on Debian-based distros::

        # apt-get install joystick

    Instructions on Arch-based distros::

        # pacman -S joyutils

    To test if your joystick is compatible, connect it to your device, check for its path (usually under
    ``/dev/input/js*``) and run::

        $ jstest /dev/input/js[n]

    Triggers:

        * :class:`platypush.message.event.joystick.JoystickConnectedEvent` when the joystick is connected.
        * :class:`platypush.message.event.joystick.JoystickDisconnectedEvent` when the joystick is disconnected.
        * :class:`platypush.message.event.joystick.JoystickStateEvent` when the state of the joystick (i.e. some of its
            axes or buttons values) changes.
        * :class:`platypush.message.event.joystick.JoystickButtonPressedEvent` when a joystick button is pressed.
        * :class:`platypush.message.event.joystick.JoystickButtonReleasedEvent` when a joystick button is released.
        * :class:`platypush.message.event.joystick.JoystickAxisEvent` when an axis value of the joystick changes.

    """

    js_axes_regex = re.compile(r'Axes:\s+(((\d+):\s*([\-\d]+)\s*)+)')
    js_buttons_regex = re.compile(r'Buttons:\s+(((\d+):\s*(on|off)\s*)+)')
    js_axis_regex = re.compile(r'^\s*(\d+):\s*([\-\d]+)\s*(.*)')
    js_button_regex = re.compile(r'^\s*(\d+):\s*(on|off)\s*(.*)')

    def __init__(self,
                 device: str = '/dev/input/js0',
                 jstest_path: str = '/usr/bin/jstest',
                 **kwargs):
        """
        :param device: Path to the joystick device (default: ``/dev/input/js0``).
        :param jstest_path: Path to the ``jstest`` executable that comes with the ``joystick`` system package
            (default: ``/usr/bin/jstest``).
        """
        super().__init__(device=device, **kwargs)

        self.device = device
        self.jstest_path = jstest_path
        self._process: Optional[subprocess.Popen] = None
        self._state: Optional[JoystickState] = None

    def _wait_ready(self):
        self.logger.info(f'Waiting for joystick device on {self.device}')

        while not self.should_stop():
            if not os.path.exists(self.device):
                time.sleep(1)

            try:
                with open(self.device, 'rb'):
                    break
            except Exception as e:
                self.logger.debug(e)
                time.sleep(0.1)
                continue

        self.bus.post(JoystickConnectedEvent(device=self.device))

    def _read_states(self):
        while not self.should_stop():
            yield self._get_state()

    def _get_state(self) -> JoystickState:
        axes = []
        buttons = []
        line = ''

        while os.path.exists(self.device) and not self.should_stop():
            ch = self._process.stdout.read(1).decode()
            if not ch:
                continue

            if ch in ['\r', '\n']:
                line = ''
                continue

            line += ch
            if line.endswith('Axes:  '):
                break

        while os.path.exists(self.device) and not self.should_stop() and len(axes) < len(self._state.axes):
            ch = ' '
            while ch == ' ':
                ch = self._process.stdout.read(1).decode()

            self._process.stdout.read(len(f'{len(axes)}'))
            value = ''

            while os.path.exists(self.device) and not self.should_stop():
                ch = self._process.stdout.read(1).decode()
                if ch == ' ':
                    if not value:
                        continue
                    break

                if ch == ':':
                    break

                value += ch

            if value:
                axes.append(int(value))

        line = ''

        while os.path.exists(self.device) and not self.should_stop():
            ch = self._process.stdout.read(1).decode()
            if not ch:
                continue

            line += ch
            if line.endswith('Buttons:  '):
                break

        while os.path.exists(self.device) and not self.should_stop() and len(buttons) < len(self._state.buttons):
            ch = ' '
            while ch == ' ':
                ch = self._process.stdout.read(1).decode()

            self._process.stdout.read(len(f'{len(buttons)}'))
            value = ''

            while os.path.exists(self.device) and not self.should_stop():
                ch = self._process.stdout.read(1).decode()
                if ch == ' ':
                    continue

                value += ch
                if value in ['on', 'off']:
                    buttons.append(value == 'on')
                    break

        return JoystickState(axes=axes, buttons=buttons)

    def _initialize(self):
        while self._process.poll() is None and \
                os.path.exists(self.device) and \
                not self.should_stop() and \
                not self._state:
            line = b''
            ch = None

            while ch not in [b'\r', b'\n']:
                ch = self._process.stdout.read(1)
                line += ch

            line = line.decode().strip()
            if not (line and line.startswith('Axes:')):
                continue

            re_axes = self.js_axes_regex.search(line)
            re_buttons = self.js_buttons_regex.search(line)

            if not (re_axes and re_buttons):
                return

            state = {
                'axes': [],
                'buttons': [],
            }

            axes = re_axes.group(1)
            while axes:
                m = self.js_axis_regex.search(axes)
                state['axes'].append(int(m.group(2)))
                axes = m.group(3)

            buttons = re_buttons.group(1)
            while buttons:
                m = self.js_button_regex.search(buttons)
                state['buttons'].append(m.group(2) == 'on')
                buttons = m.group(3)

            self._state = JoystickState(**state)

    def _process_state(self, state: JoystickState):
        diff = self._state - state
        if not diff:
            return

        self.bus.post(JoystickStateEvent(device=self.device, **state.__dict__))

        for button, pressed in diff.get('buttons', {}).items():
            evt_class = JoystickButtonPressedEvent if pressed else JoystickButtonReleasedEvent
            self.bus.post(evt_class(device=self.device, button=button))

        for axis, value in diff.get('axes', {}).items():
            self.bus.post(JoystickAxisEvent(device=self.device, axis=axis, value=value))

        self._state = state

    def run(self):
        super().run()

        try:
            while not self.should_stop():
                self._wait_ready()

                with subprocess.Popen(
                        [self.jstest_path, '--normal', self.device],
                        stdout=subprocess.PIPE) as self._process:
                    self.logger.info('Device opened')
                    self._initialize()

                    if self._process.poll() is not None:
                        break

                    for state in self._read_states():
                        if self._process.poll() is not None or not os.path.exists(self.device):
                            self.logger.warning(f'Connection to {self.device} lost')
                            self.bus.post(JoystickDisconnectedEvent(self.device))
                            break

                        self._process_state(state)
        finally:
            self._process = None


# vim:sw=4:ts=4:et:



import logging
import os
import sys

logging.basicConfig(level=logging.INFO, stream=sys.stdout)
test_dir = os.path.abspath(os.path.dirname(__file__))
sys.path.insert(0, os.path.abspath(os.path.join(test_dir, '..')))



import requests
from typing import Optional
from urllib.parse import urljoin, urlencode
from platypush.backend.http.app.utils import get_local_base_url

from platypush.context import get_backend
from platypush.plugins import action
from platypush.plugins.tts import TtsPlugin
from platypush.schemas.tts.mimic3 import Mimic3VoiceSchema


class TtsMimic3Plugin(TtsPlugin):
    """
    TTS plugin that uses the `Mimic3 webserver
    <https://github.com/MycroftAI/mimic3>`_ provided by `Mycroft
    <https://mycroft.ai/>`_ as a text-to-speech engine.

    The easiest way to deploy a Mimic3 instance is probably via Docker:

    .. code-block:: bash

        $ mkdir -p "$HOME/.local/share/mycroft/mimic3"
        $ chmod a+rwx "$HOME/.local/share/mycroft/mimic3"
        $ docker run --rm \
            -p 59125:59125 \
            -v "%h/.local/share/mycroft/mimic3:/home/mimic3/.local/share/mycroft/mimic3" \
            'mycroftai/mimic3'

    Requires:

        * At least a *media plugin* (see
          :class:`platypush.plugins.media.MediaPlugin`) enabled/configured -
          used for speech playback.
        * The ``http`` backend (:class:`platypush.backend.http.HttpBackend`)
          enabled - used for proxying the API calls.

    """

    def __init__(
        self,
        server_url: str,
        voice: str = 'en_UK/apope_low',
        media_plugin: Optional[str] = None,
        player_args: Optional[dict] = None,
        **kwargs
    ):
        """
        :param server_url: Base URL of the web server that runs the Mimic3 engine.
        :param voice: Default voice to be used (default: ``en_UK/apope_low``).
            You can get a full list of the voices available on the server
            through :meth:`.voices`.
        :param media_plugin: Media plugin to be used for audio playback. Supported:

            - ``media.gstreamer``
            - ``media.omxplayer``
            - ``media.mplayer``
            - ``media.mpv``
            - ``media.vlc``

        :param player_args: Optional arguments that should be passed to the player plugin's
            :meth:`platypush.plugins.media.MediaPlugin.play` method.
        """
        super().__init__(media_plugin=media_plugin, player_args=player_args, **kwargs)

        self.server_url = server_url
        self.voice = voice

    @action
    def say(
        self,
        text: str,
        server_url: Optional[str] = None,
        voice: Optional[str] = None,
        player_args: Optional[dict] = None,
    ):
        """
        Say some text.

        :param text: Text to say.
        :param server_url: Default ``server_url`` override.
        :param voice: Default ``voice`` override.
        :param player_args: Default ``player_args`` override.
        """
        server_url = server_url or self.server_url
        voice = voice or self.voice
        player_args = player_args or self.player_args
        http = get_backend('http')
        assert http, 'http backend not configured'
        assert self.media_plugin, 'No media plugin configured'

        url = (
            urljoin(get_local_base_url(), '/tts/mimic3/say')
            + '?'
            + urlencode(
                {
                    'text': text,
                    'server_url': server_url,
                    'voice': voice,
                }
            )
        )

        self.media_plugin.play(url, **player_args)

    @action
    def voices(self, server_url: Optional[str] = None):
        """
        List the voices available on the server.

        :param server_url: Default ``server_url`` override.
        :return: .. schema:: tts.mimic3.Mimic3VoiceSchema(many=True)
        """
        server_url = server_url or self.server_url
        rs = requests.get(urljoin(server_url, '/api/voices'))
        rs.raise_for_status()
        return Mimic3VoiceSchema().dump(rs.json(), many=True)


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import datetime
import requests
from typing import Optional

from platypush.plugins import Plugin, action
from platypush.plugins.calendar import CalendarInterface


class CalendarIcalPlugin(Plugin, CalendarInterface):
    """
    iCal calendars plugin. Interact with remote calendars in iCal format.

    Requires:

        * **icalendar** (``pip install icalendar``)

    """

    def __init__(self, url, *args, **kwargs):
        """
        :param url: iCal URL to parse
        :type url: str
        """
        super().__init__(*args, **kwargs)
        self.url = url

    @staticmethod
    def _convert_timestamp(event, attribute: str) -> Optional[str]:
        t = event.get(attribute)
        if not t:
            return

        if type(t.dt) == datetime.date:
            return (
                datetime.datetime(
                    t.dt.year, t.dt.month, t.dt.day, tzinfo=datetime.timezone.utc
                ).isoformat()
            )

        return (
                datetime.datetime.utcfromtimestamp(t.dt.timestamp())
                .replace(tzinfo=datetime.timezone.utc).isoformat()
        )

    @classmethod
    def _translate_event(cls, event):
        return {
            'id': str(event.get('uid')) if event.get('uid') else None,
            'kind': 'calendar#event',
            'summary': str(event.get('summary')) if event.get('summary') else None,
            'description': str(event.get('description')) if event.get('description') else None,
            'status': str(event.get('status')).lower() if event.get('status') else None,
            'responseStatus': str(event.get('partstat')).lower() if event.get('partstat') else None,
            'location': str(event.get('location')) if event.get('location') else None,
            'htmlLink': str(event.get('url')) if event.get('url') else None,
            'organizer': {
                'email': str(event.get('organizer')).replace('MAILTO:', ''),
                'displayName': event.get('organizer').params.get('cn')
            } if event.get('organizer') else None,

            'created': cls._convert_timestamp(event, 'created'),
            'updated': cls._convert_timestamp(event, 'last-modified'),
            'start': {
                'dateTime': cls._convert_timestamp(event, 'dtstart'),
                'timeZone': 'UTC',
            },

            'end': {
                'dateTime': cls._convert_timestamp(event, 'dtend'),
                'timeZone': 'UTC',
            },
        }

    @action
    def get_upcoming_events(self, max_results=10, only_participating=True):
        """
        Get the upcoming events. See
        :func:`~platypush.plugins.calendar.CalendarPlugin.get_upcoming_events`.
        """

        from icalendar import Calendar

        events = []
        response = requests.get(self.url)
        assert response.ok, \
            "HTTP error while getting events from {}: {}".format(self.url, response.text)

        calendar = Calendar.from_ical(response.text)
        for event in calendar.walk():
            if event.name != 'VEVENT':
                continue  # Not an event

            event = self._translate_event(event)

            if (
                    event['status'] != 'cancelled'
                    and event['end'].get('dateTime')
                    and event['end']['dateTime'] >= datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()
                    and (
                    (only_participating
                     and event.get('responseStatus') in [None, 'accepted', 'tentative'])
                    or not only_participating)
            ):
                events.append(event)

        return events

# vim:sw=4:ts=4:et:
